#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on 2021.03.19
@author: zoharslong
"""
from pyzohar.sub_slt_bsc.bsz import stz, lsz, dtz, dcz
from pyzohar.sub_slt_bsc.dfz import dfz

__all__ = ['stz', 'lsz', 'dtz', 'dcz', 'dfz']
print('info: basic slot inserted.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 22 16:30:38 2019
@author: zoharslong
"""
from platform import system as chk_sys
from os.path import realpath as os_realpath
from os import listdir as os_listdir
from pyzohar.sub_slt_bsc.bsz import stz, lsz, dcz, dtz
from pyzohar.sub_slt_bsc.dfz import dfz


chk_sys = chk_sys()
lst_xcd = ['samples', '__pycache__', '__init__.py']  # list for exclude
lst_slt = [i for i in os_listdir(os_realpath(__file__).replace('__init__.py', '')) if i not in lst_xcd]
__version__ = '0.1.12'
__doc__ = """
pyzohar
=====================================================================
** pyzohar ** is a Python package for data manipulation
** main features
  - basic data type pre-processing.
  - pandas.DataFrame manipulating.
  - data input and output.
  - slot for fast modelling.
  - slot for api and web spider.
"""
__all__ = [
    'stz',
    'lsz',
    'dcz',
    'dtz',
    'dfz',
]   # https://www.cnblogs.com/tp1226/p/8453854.html

print('info: pyzohar_%s initiation succeeded.' % __version__)
print('info: slots stand by: %s' % lst_slt)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Mon May 31 10:04:01 2021
ppt generating operation.
requirements = [
  'python-pptx>=0.6.19',
]
@author: zoharslong
"""
from pyzohar.sub_slt_bsc.bsz import lsz
from pyzohar.sub_slt_bsc.dfz import dfz
from pandas import DataFrame, isnull
from pptx import Presentation
from os.path import join as os_join, exists as os_exists
from pptx.enum.shapes import MSO_SHAPE
from pptx.enum.text import MSO_ANCHOR, MSO_AUTO_SIZE, PP_ALIGN
from pptx.dml.color import RGBColor
from pptx.util import Cm, Pt, Inches
from pptx.presentation import Presentation as typ_presentation
from pptx.chart.data import CategoryChartData
from pptx.enum.chart import XL_CHART_TYPE, XL_TICK_MARK, XL_LEGEND_POSITION, XL_LABEL_POSITION
dct_plt = {
    'blk': (0, 0, 0),  # 黑
    'wht': (255, 255, 255),  # 白
    'gry': (191, 191, 191),  # 灰
    'gry_dep': (127, 127, 127),  # 深灰
    'red_lgt': (225, 178, 172),  # 浅红
    'red_nrm': (209, 136, 126),  # 红
    'red_dep': (167, 91, 81),  # 深红
}  # 调色板


class ppz_nit(dfz):
    """
    self.lcn = {fld, fls, ppt:{mtr,sld,...}}
    """
    dct_util = {
        'cm': Cm,
        'pt': Pt,
        'nch': Inches
    }              # 尺度单位
    dct_anchor = {
        'bottom': MSO_ANCHOR.BOTTOM,
        'middle': MSO_ANCHOR.MIDDLE,
        'center': MSO_ANCHOR.MIDDLE,
        'top': MSO_ANCHOR.TOP
    }            # 纵向对齐
    dct_align = {
        'justify': PP_ALIGN.JUSTIFY,  # 两端
        'left': PP_ALIGN.LEFT,
        'right': PP_ALIGN.RIGHT,
        'center': PP_ALIGN.CENTER,  # 居中
        'middle': PP_ALIGN.CENTER,  # 居中
    }             # 横向对齐
    dct_chart_type = {
        'bar': XL_CHART_TYPE.COLUMN_CLUSTERED,
        'barchart': XL_CHART_TYPE.COLUMN_CLUSTERED,
        'piechart': XL_CHART_TYPE.PIE,
    }
    dct_legend_position = {
        'rightout': [XL_LEGEND_POSITION.RIGHT, False],
        'bottomout': [XL_LEGEND_POSITION.BOTTOM, False],
        'rightin': [XL_LEGEND_POSITION.RIGHT, True],
        'bottomin': [XL_LEGEND_POSITION.BOTTOM, True],
    }   # 图例位置
    dct_label_position = {
        'inside_base': XL_LABEL_POSITION.INSIDE_BASE,
        'inside_end': XL_LABEL_POSITION.INSIDE_END,
        'outside_end': XL_LABEL_POSITION.OUTSIDE_END,
        'best_fit': XL_LABEL_POSITION.BEST_FIT,
        'center': XL_LABEL_POSITION.CENTER,
        'above': XL_LABEL_POSITION.ABOVE,
        'right': XL_LABEL_POSITION.RIGHT,
    }    # 标签位置
    dct_mtr = {
        'wdh': 9144000,
        'wtd': '16:10',
        'utl': 'cm',
        'fontName': 'Microsoft YaHei UI',
        'fontColor': (64, 64, 64),
        'fontSize': Pt(12),
        'fontSize_label': Pt(10),
        'fontSize_label_2': Pt(8),
        'fontBold': False,
        'fontItalic': False,
        'anchor': 'middle',  # 纵向对齐 - 居中
        'align': 'justify',  # 横向对齐 - 两端
        'line_spacing': 1.15,
        'space_before': Pt(3),
        'space_after': Pt(3),
        'mtr': {
            'up3down1': [
                {
                    'typ': 'textbox', 'tip': '顶栏总结',
                    'left': 1.4 / 2, 'top': 0.875 / 2, 'width': 25.4 - 1.4 - 4.5, 'height': 1.5,
                    'anchor': 'middle', 'align': 'justify',
                },  # 顶栏总结 - 需要{paragraph:[{ctt}]}
                {
                    'tip': '上半内容1/3',
                    'left': 1.4 / 2,
                    'top': 0.875 / 2 + 1.5 + 1,
                    'width': (25.4 - 1.4-1-2)/3,
                    'height': (15.875 - (0.875 + 1.5 + 1))/2 - 0.25,
                },  # 上半内容 - 需要{typ, ..}
                {
                    'tip': '上半内容2/3',
                    'left': 1.4/2+(25.4-1.4-1-2)/3+0.5,
                    'top': 0.875 / 2 + 1.5 + 1,
                    'width': (25.4 - 1.4-1-2)/3,
                    'height': (15.875 - (0.875 + 1.5 + 1))/2 - 0.25,
                },  # 上半内容 - 需要{typ, ..}
                {
                    'tip': '上半内容3/3',
                    'left': 1.4/2+(25.4-1.4-1-2)/3+0.5+(25.4-1.4-1-2)/3+0.5,
                    'top': 0.875 / 2 + 1.5 + 1,
                    'width': (25.4 - 1.4-1-2)/3 + 2,
                    'height': (15.875 - (0.875 + 1.5 + 1))/2 - 0.25,
                },  # 上半内容 - 需要{typ, ..}
                {
                    'tip': '下半内容',
                    'left': 1.4 / 2, 'top': 0.875 / 2 + 1.5 + 1 + (15.875 - (0.875 + 1.5 + 1))/2 + 0.25,
                    'width': (25.4 - 1.4),
                    'height': (15.875 - (0.875 + 1.5 + 1))/2 - 0.25,
                },  # 下半内容 - 需要{typ, ..}
                {
                    'typ': 'picture', 'tip': '右上图标',
                    'left': 25.4 - 4.5, 'top': 0.875 / 2, 'height': 1.5, 'width': 4.5,
                },  # 右上图标 - 需要图片地址的{fld, fls}
                {
                    'typ': 'line', 'tip': '隔断横线',
                    'left': 1.4 / 2 + 3, 'top': 0.875 / 2 + 1.5 + 1 + (15.875 - (0.875 + 1.5 + 1))/2 - 0.001,
                    'width': 25.4 - 1.4 - 6,
                    'height': 0.002,
                    'color': (225, 225, 225),
                },  # 隔断中线 - 可省略
            ],  # 上三分下一块
            'full': [
                {
                    'typ': 'textbox', 'tip': '顶栏总结',
                    'left': 1.4 / 2, 'top': 0.875 / 2, 'width': 25.4 - 1.4 - 4.5, 'height': 1.5,
                    'anchor': 'middle', 'align': 'justify',
                },  # 顶栏总结 - 需要{paragraph:[{ctt}]}
                {
                    'tip': '左半内容',
                    'left': 1.4 / 2, 'top': 0.875 / 2 + 1.5 + 1, 'width': (25.4 - 1.4),
                    'height': 15.875 - (0.875 + 1.5 + 1),
                },  # 全页内容 - 需要{typ, ..}
                {
                    'typ': 'picture', 'tip': '右上图标',
                    'left': 25.4 - 4.5, 'top': 0.875 / 2, 'height': 1.5, 'width': 4.5,
                },  # 右上图标 - 需要图片地址的{fld, fls}
            ],    # 整版呈现布局[title, full, logo]
            'middle': [
                {
                    'typ': 'textbox', 'tip': '顶栏总结',
                    'left': 1.4 / 2, 'top': 0.875 / 2, 'width': 25.4 - 1.4 - 4.5, 'height': 1.5,
                    'anchor': 'middle', 'align': 'justify',
                },  # 顶栏总结 - 需要{paragraph:[{ctt}]}
                {
                    'tip': '左半内容',
                    'left': 1.4 / 2, 'top': 0.875 / 2 + 1.5 + 1, 'width': (25.4 - 1.4) / 2 - 0.25,
                    'height': 15.875 - (0.875 + 1.5 + 1),
                },  # 左半内容 - 需要{typ, ..}
                {
                    'tip': '右半内容',
                    'left': 25.4 / 2 + 0.25, 'top': 0.875 / 2 + 1.5 + 1, 'width': (25.4 - 1.4) / 2 - 0.25,
                    'height': 15.875 - (0.875 + 1.5 + 1),
                },  # 右半内容 - 需要{typ, ..}
                {
                    'typ': 'picture', 'tip': '右上图标',
                    'left': 25.4 - 4.5, 'top': 0.875 / 2, 'height': 1.5, 'width': 4.5,
                },  # 右上图标 - 需要图片地址的{fld, fls}
                {
                    'typ': 'line', 'tip': '隔断中线',
                    'left': 25.4 / 2 - 0.001, 'top': 0.875 / 2 + 1.5 + 1 + 3, 'width': 0.002,
                    'height': 15.875 - (0.875 + 1.5 + 1) - 6,
                    'color': (225, 225, 225),
                },  # 隔断中线 - 可省略
            ],  # 中位分割布局[title, left, right, logo, splitLine]
            'center': [
                {
                    'typ': 'textbox', 'tip': '顶栏总结',
                    'left': 1.4 / 2, 'top': 0.875 / 2, 'width': 25.4 - 1.4 - 4.5, 'height': 1.5,
                    'anchor': 'middle', 'align': 'justify',
                },  # 顶栏总结 - 需要{paragraph:[{ctt}]}
                {
                    'tip': '上半内容',
                    'left': 1.4 / 2, 'top': 0.875 / 2 + 1.5 + 1,
                    'width': (25.4 - 1.4),
                    'height': (15.875 - (0.875 + 1.5 + 1))/2 - 0.25,
                },  # 上半内容 - 需要{typ, ..}
                {
                    'tip': '下半内容',
                    'left': 1.4 / 2, 'top': 0.875 / 2 + 1.5 + 1 + (15.875 - (0.875 + 1.5 + 1))/2 + 0.25,
                    'width': (25.4 - 1.4),
                    'height': (15.875 - (0.875 + 1.5 + 1))/2 - 0.25,
                },  # 下半内容 - 需要{typ, ..}
                {
                    'typ': 'picture', 'tip': '右上图标',
                    'left': 25.4 - 4.5, 'top': 0.875 / 2, 'height': 1.5, 'width': 4.5,
                },  # 右上图标 - 需要图片地址的{fld, fls}
                {
                    'typ': 'line', 'tip': '隔断横线',
                    'left': 1.4 / 2 + 3, 'top': 0.875 / 2 + 1.5 + 1 + (15.875 - (0.875 + 1.5 + 1))/2 - 0.001,
                    'width': 25.4 - 1.4 - 6,
                    'height': 0.002,
                    'color': (225, 225, 225),
                },  # 隔断中线 - 可省略
            ],
            'bottom': [],
            'left': [
                {
                    'typ': 'textbox',
                    'tip': '顶栏总结',
                    'left': 1.4 / 2,
                    'top': 0.875 / 2,
                    'width': 25.4 - 1.4 - 4.5,
                    'height': 1.5,
                    'anchor': 'middle',
                    'align': 'justify',
                },  # 顶栏总结 - 需要{paragraph:[{ctt}]}
                {
                    'tip': '左半内容',
                    'left': 1.4 / 2,
                    'top': 0.875 / 2 + 1.5 + 1,
                    'width': (25.4 - 1.4) / 3*2 - 0.25,
                    'height': 15.875 - (0.875 + 1.5 + 1),
                },  # 左半内容 - 需要{typ, ..}
                {
                    'tip': '右半内容',
                    'left': 1.4 / 2 + (25.4-1.4) / 3*2 + 0.25,
                    'top': 0.875 / 2 + 1.5 + 1,
                    'width': (25.4 - 1.4) / 3*1 - 0.25,
                    'height': 15.875 - (0.875 + 1.5 + 1),
                },  # 右半内容 - 需要{typ, ..}
                {
                    'typ': 'picture', 'tip': '右上图标',
                    'left': 25.4 - 4.5, 'top': 0.875 / 2, 'height': 1.5, 'width': 4.5,
                },  # 右上图标 - 需要图片地址的{fld, fls}
                {
                    'typ': 'line', 'tip': '隔断中线',
                    'left': 1.4/2 + (25.4-1.4) / 3*2 - 0.001,
                    'top': 0.875 / 2 + 1.5 + 1 + 3,
                    'width': 0.002,
                    'height': 15.875 - (0.875 + 1.5 + 1) - 6,
                    'color': (225, 225, 225),
                },  # 隔断中线 - 可省略
            ],
            'right': [
                {
                    'typ': 'textbox',
                    'tip': '顶栏总结',
                    'left': 1.4 / 2,
                    'top': 0.875 / 2,
                    'width': 25.4 - 1.4 - 4.5,
                    'height': 1.5,
                    'anchor': 'middle',
                    'align': 'justify',
                },  # 顶栏总结 - 需要{paragraph:[{ctt}]}
                {
                    'tip': '左半内容',
                    'left': 1.4 / 2,
                    'top': 0.875 / 2 + 1.5 + 1,
                    'width': (25.4 - 1.4) / 3*1 - 0.25,
                    'height': 15.875 - (0.875 + 1.5 + 1),
                },  # 左半内容 - 需要{typ, ..}
                {
                    'tip': '右半内容',
                    'left': 1.4 / 2 + (25.4-1.4) / 3*1 + 0.25,
                    'top': 0.875 / 2 + 1.5 + 1,
                    'width': (25.4 - 1.4) / 3*2 - 0.25,
                    'height': 15.875 - (0.875 + 1.5 + 1),
                },  # 右半内容 - 需要{typ, ..}
                {
                    'typ': 'picture', 'tip': '右上图标',
                    'left': 25.4 - 4.5, 'top': 0.875 / 2, 'height': 1.5, 'width': 4.5,
                },  # 右上图标 - 需要图片地址的{fld, fls}
                {
                    'typ': 'line', 'tip': '隔断中线',
                    'left': 1.4/2 + (25.4-1.4) / 3*1 - 0.001,
                    'top': 0.875 / 2 + 1.5 + 1 + 3,
                    'width': 0.002,
                    'height': 15.875 - (0.875 + 1.5 + 1) - 6,
                    'color': (225, 225, 225),
                },  # 隔断中线 - 可省略
            ],
        },  # 可以直接在slide调用的布局, 通过lcn.ppt.sld.mtr指定
        'tbl': {
            'mrg': [],
            'cll_fit': {},
            'cll_fit_clm': {},
            'anchor': 'middle',
            'align': 'middle',
            'color': (225, 225, 225),
            'fontName': 'Microsoft YaHei',
            'fontColor': (0, 0, 0),
            'fontSize': 12,  # fixed Pt
            'fontBold': False,
            'fontItalic': False,
            'margin_left': 0.15,     # 跟mtr.utl
            'margin_right': 0.15,    # 跟mtr.utl
        },
    }               # 母版的默认参数
    dct_plt = dct_plt             # 调色板

    def __init__(self, dts=None, lcn=None, *, spr=False, cvr=True, ppt=None):
        super(ppz_nit, self).__init__(dts, lcn, spr=spr)
        self.__ppt = None
        self.ppt = ppt
        self.ppt_nit(cvr=cvr)

    @property
    def ppt(self):
        """
        self.ppt.
        :return: self.__ppt
        """
        return self.__ppt

    @ppt.setter
    def ppt(self, ppt):
        """
        set self.__ppt in self.ppt.
        :param ppt: main object from python-pptx
        :return: None
        """
        if type(ppt) in [typ_presentation, type(None)]:
            self.__ppt = ppt
        else:
            raise TypeError('info: ppt\'s type %s is not available.' % type(ppt))

    def ppt_nit_chk_lcn(self):
        if 'ppt' not in self.lcn.keys():
            self.lcn.update({'ppt': {'mtr': self.dct_mtr}})
        elif 'mtr' not in self.lcn['ppt'].keys():
            self.lcn['ppt'].update({'mtr': self.dct_mtr})
        else:
            dct_tmp = self.dct_mtr.copy()
            dct_tmp.update(self.lcn['ppt']['mtr'])
            self.lcn['ppt']['mtr'].update(dct_tmp)

    def ppt_nit(self, *, cvr=True):
        """
        pptx initiation,
        create a new empty self.ppt or get ppt from file
        :param cvr: create a new empty self.ppt to cover the old one in local file, default True
        """
        fls = os_join(self.lcn['fld'], self.lcn['fls']) if \
            os_exists(os_join(self.lcn['fld'], self.lcn['fls'])) and cvr is False else None
        self.ppt = Presentation(fls) if \
            self.ppt is None or type(self.ppt) not in [typ_presentation] or cvr is True else self.ppt
        self.ppt_nit_chk_lcn()

    def ppt_siz(self, wdh=None, wtd=None):
        """
        pptx's size, width and height default [10Inches, 25.4Cm, 9144000], [6.25Inches, 15.875Cm, 5715000]
        :param wdh: width
        :param wtd: width to height, 宽高比
        """
        if 'wdh' in self.lcn['ppt']['mtr'].keys():
            self.lcn['ppt']['mtr']['wdh'] = wdh = wdh if \
                wdh is not None and self.lcn['ppt']['mtr']['wdh'] is None else self.lcn['ppt']['mtr']['wdh']
        else:
            self.lcn['ppt']['mtr']['wdh'] = wdh
        if 'wtd' in self.lcn['ppt']['mtr'].keys():
            self.lcn['ppt']['mtr']['wtd'] = wtd = wtd if \
                wtd is not None and self.lcn['ppt']['mtr']['wtd'] is None else self.lcn['ppt']['mtr']['wtd']
        else:
            self.lcn['ppt']['mtr']['wtd'] = wtd
        self.ppt.slide_width = wdh
        self.ppt.slide_height = int(wdh / float(wtd.split(':')[0]) * float(wtd.split(':')[1]))

    def sld_mtr(self, ndx_sld):
        """"""
        if 'mtr' not in self.lcn['ppt'][ndx_sld] or self.lcn['ppt'][ndx_sld]['mtr'] is None:
            pass
        else:
            print('info: mother mask %s synchronization' % self.lcn['ppt'][ndx_sld]['mtr'])
            mtr = self.lcn['ppt'][ndx_sld]['mtr']
            n = 0
            for dct_i in self.dct_mtr['mtr'][mtr]:
                dct_i_tmp = dct_i.copy()
                try:
                    dct_i_tmp.update(self.lcn['ppt'][ndx_sld]['shp'][n])
                    self.lcn['ppt'][ndx_sld]['shp'][n].update(dct_i_tmp)
                except IndexError:
                    self.lcn['ppt'][ndx_sld]['shp'].append(dct_i_tmp)
                finally:
                    n += 1


class tblMixin(ppz_nit):

    def __init__(self, dts=None, lcn=None, *, spr=False, cvr=True):
        super(tblMixin, self).__init__(dts, lcn, spr=spr, cvr=cvr)

    def tbl_mtr(self, dct_shp):
        dct_tmp = self.dct_mtr['tbl'].copy()
        dct_tmp.update(dct_shp)
        dct_shp.update(dct_tmp)
        if 'row' not in dct_shp:
            dct_shp['row'], dct_shp['clm'] = DataFrame(dct_shp['cll']).shape

    def fit_mtr(self, dct_shp, tgt_fit='cll_fit', flt_len=None, *, dct_fit=None):
        """
        对dct_shp.cll_fit进行批量处理
        """
        flt_len = dct_shp['clm'] if flt_len is None else flt_len
        dct_shp[tgt_fit] = dct_fit if dct_fit is not None else dct_shp[tgt_fit]
        dct_shp[tgt_fit] = self.dct_mtr['tbl'][tgt_fit] if tgt_fit not in dct_shp else dct_shp[tgt_fit]
        for i_fit in dct_shp[tgt_fit]:
            for i_row in range(len(dct_shp[tgt_fit][i_fit])):
                if type(dct_shp[tgt_fit][i_fit][i_row]) not in [list]:
                    dct_shp[tgt_fit][i_fit][i_row] = lsz([dct_shp[tgt_fit][i_fit][i_row]]).cpy_tal(flt_len, rtn=True)

    def _it_cll(self, dct_shp, str_tgt='cll_fit'):
        """
        fit cell format operation part
        :param str_tgt: 选择是对行还是列进行批量修改, in ['cll_fit','cll_fit_clm'] for [行, 列]
        """
        dct_shp[str_tgt] = self.dct_mtr['tbl'][str_tgt] if str_tgt not in dct_shp else dct_shp[str_tgt]
        for i_fit in dct_shp[str_tgt]:
            for i_row in range(len(dct_shp[str_tgt][i_fit])):
                for i_clm in range(len(dct_shp[str_tgt][i_fit][i_row])):
                    try:
                        if type(dct_shp['cll'][i_row][i_clm]) not in [dict]:
                            dct_shp['cll'][i_row][i_clm] = {'ctt': dct_shp['cll'][i_row][i_clm]}
                            dct_shp['cll'][i_row][i_clm]['ctt'] = str(dct_shp['cll'][i_row][i_clm]['ctt']) if \
                                'ctt' in dct_shp['cll'][i_row][i_clm] and \
                                not isnull(dct_shp['cll'][i_row][i_clm]['ctt']) else ''
                        if i_fit not in dct_shp['cll'][i_row][i_clm]:
                            dct_shp['cll'][i_row][i_clm][i_fit] = dct_shp[str_tgt][i_fit][i_row][i_clm]
                    except IndexError:
                        break

    def fit_cll(self, dct_shp):
        """
        比较方便地更新self.lcn.ppt.sld.shp.cll
        """
        # 行单元格格式批量处理
        self.fit_mtr(dct_shp, 'cll_fit', dct_shp['clm'])
        self._it_cll(dct_shp, 'cll_fit')
        # 列单元格格式批量修改
        self.fit_mtr(dct_shp, 'cll_fit_clm', dct_shp['row'])
        dct_shp['cll'] = list(map(list, zip(*dct_shp['cll'])))  # 列格式批量处理, 先转置, 再转置还原
        self._it_cll(dct_shp, 'cll_fit_clm')  # 使用dct_shp.cll_fit_clm中的数据进行批量格式赋予
        dct_shp['cll'] = list(map(list, zip(*dct_shp['cll'])))

    def gnr_mrg(self, dct_shp):
        """
        单元格合并, 参数使用dct_shp['mrg']
        """
        dct_shp['mrg'] = self.dct_mtr['tbl']['mrg'] if 'mrg' not in dct_shp else dct_shp['mrg']  # meaningless
        for i_mrg in dct_shp['mrg']:
            dct_shp['obj_tbl'].cell(i_mrg[0], i_mrg[1]).merge(dct_shp['obj_tbl'].cell(i_mrg[2], i_mrg[3]))

    def gnr_cll(self, dct_shp):
        """"""
        dct_shp['obj_tbl'].first_row = False  # 第一行特殊化
        dct_shp['obj_tbl'].first_col = False  # 第一列特殊化
        lst_cll = lsz(dct_shp['cll']).nfd(rtn=True)  # 单元格展开成一行方便后续的赋值操作
        n = 0
        for i_cll in dct_shp['obj_tbl'].iter_cells():
            lst_cll[n] = {'ctt': lst_cll[n]} if type(lst_cll[n]) not in [dict] else lst_cll[n]
            lst_cll[n]['ctt'] = str(lst_cll[n]['ctt']) if 'ctt' in lst_cll[n] and not isnull(lst_cll[n]['ctt']) else ''
            bkg_rgb = lst_cll[n]['color'] if 'color' in lst_cll[n] else dct_shp['color']
            i_cll.fill.solid()
            i_cll.fill.fore_color.rgb = RGBColor(bkg_rgb[0], bkg_rgb[1], bkg_rgb[2])  # 单元格背景色
            i_cll.margin_left = self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['margin_left'])
            i_cll.margin_right = self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['margin_right'])
            i_cll.vertical_anchor = self.dct_anchor[dct_shp['anchor']]  # 纵向对齐
            prh = i_cll.text_frame.paragraphs[0]                # 进入单元格文本处理
            prh.alignment = self.dct_align[dct_shp['align']]    # 横向对齐
            prh.font.size = Pt(lst_cll[n]['fontSize']) if 'fontSize' in lst_cll[n] else Pt(dct_shp['fontSize'])
            prh.font.name = lst_cll[n]['fontName'] if 'fontName' in lst_cll[n] else dct_shp['fontName']
            prh.font.bold = lst_cll[n]['fontBold'] if 'fontBold' in lst_cll[n] else dct_shp['fontBold']
            prh.font.italic = lst_cll[n]['fontItalic'] if 'fontItalic' in lst_cll[n] else dct_shp['fontItalic']
            fnt_rgb = lst_cll[n]['fontColor'] if 'fontColor' in lst_cll[n] else dct_shp['fontColor']
            prh.font.color.rgb = RGBColor(fnt_rgb[0], fnt_rgb[1], fnt_rgb[2])
            prh.text = lst_cll[n]['ctt'] if 'ctt' in lst_cll[n] else ''
            n += 1

    def gnr_tbl(self, ndx_sld, dct_shp):
        """"""
        self.tbl_mtr(dct_shp)
        dct_shp['obj_tbl'] = self.lcn['ppt'][ndx_sld]['obj_shp'].add_table(
            dct_shp['row'],
            dct_shp['clm'],
            self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['left']),
            self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['top']),
            self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['width']),
            self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['height']),
        ).table
        self.fit_cll(dct_shp)
        self.gnr_cll(dct_shp)
        self.gnr_mrg(dct_shp)


class pltMixin(ppz_nit):

    def __init__(self, dts=None, lcn=None, *, spr=False, cvr=True):
        super(pltMixin, self).__init__(dts, lcn, spr=spr, cvr=cvr)

    def gnr_ttl(self, dct_shp):
        """
        generate title, needs 'ttl' in dct_shp
        """
        dct_m = self.lcn['ppt']['mtr']
        if 'ttl' in dct_shp.keys():
            dct_shp['obj_plt'].has_title = True  # 图表是否含有标题，默认为False
            dct_shp['obj_plt'].chart_title.text_frame.clear()  # 清除原标题
            prh = dct_shp['obj_plt'].chart_title.text_frame.paragraphs[0]  # 添加一行新标题
            run = prh.add_run()
            run.text = dct_shp['ttl']
            run.font.size = Pt(dct_shp['titleFontSize']) if 'titleFontSize' in dct_shp else dct_m['fontSize']
            run.font.color.rgb = RGBColor(*self.dct_plt['gry_dep'])
            run.font.name = dct_m['fontName']
            run.font.bold = dct_m['fontBold']
        else:
            dct_shp['obj_plt'].has_title = False

    def gnr_lgd(self, dct_shp):
        """
        generate legend, needs 'lgd' in dct_shp, ['rightout','rightin','leftout/in','bottomout/in','topout/in']
        """
        dct_m = self.lcn['ppt']['mtr']
        if 'lgd' in dct_shp.keys():
            dct_shp['obj_plt'].has_legend = True
            dct_shp['obj_plt'].legend.position = self.dct_legend_position[dct_shp['lgd'].lower()][0]
            dct_shp['obj_plt'].legend.include_in_layout = self.dct_legend_position[dct_shp['lgd']][1]
            dct_shp['obj_plt'].legend.font.name = dct_m['fontName']
            dct_shp['obj_plt'].legend.font.size = dct_m['fontSize_label']
            dct_shp['obj_plt'].legend.font.bold = False
            dct_shp['obj_plt'].legend.font.italic = False
            dct_shp['obj_plt'].legend.font.color.rgb = RGBColor(*self.dct_plt['gry_dep'])
        else:
            dct_shp['obj_plt'].has_legend = False

    def gnr_axs(self, dct_shp):
        """
        generate axis in category and value
        """
        try:
            dct_m = self.lcn['ppt']['mtr']
            for i_axs in [dct_shp['obj_plt'].category_axis, dct_shp['obj_plt'].value_axis]:
                i_axs.has_major_gridlines = False
                i_axs.has_minor_gridlines = False
                i_axs.minor_tick_mark = XL_TICK_MARK.NONE
                i_axs.major_tick_mark = XL_TICK_MARK.NONE
                i_axs.tick_labels.font.name = dct_m['fontName']
                i_axs.tick_labels.font.size = dct_m['fontSize_label'] if \
                    i_axs in [dct_shp['obj_plt'].category_axis] else dct_m['fontSize_label_2']
                i_axs.tick_labels.font.bold = False
                i_axs.tick_labels.font.italic = False
                i_axs.tick_labels.font.color.rgb = RGBColor(*self.dct_plt['gry_dep'])
                i_axs.format.line.color.rgb = RGBColor(*self.dct_plt['gry'])
                # i_axs.maximum_scale = 100.0  # 纵坐标最大值
                # i_axs.minimum_scale = 0.0  # 纵坐标最小值
        except ValueError:
            print('info: no axis in this type of chart.')

    def gnr_lbl(self, dct_shp):
        """
        generate label in a plot
        bar charts in ['outside_end', 'inside_base'] for [柱外顶端, 柱内贴进横轴]
        """
        dct_m = self.lcn['ppt']['mtr']
        plt = dct_shp['obj_plt'].plots[0]
        if 'lbl' in dct_shp:
            plt.has_data_labels = True
            plt.data_labels.font.name = dct_m['fontName']
            plt.data_labels.font.size = dct_m['fontSize_label']
            plt.data_labels.font.color.rgb = RGBColor(*self.dct_plt['gry_dep'])
            plt.data_labels.position = self.dct_label_position[dct_shp['lbl'].lower()]
        else:
            plt.has_data_labels = False

    def gnr_plt(self, ndx_sld, dct_shp):
        """"""
        chart_data = CategoryChartData()
        chart_data.categories = dct_shp['clm']
        for i in dct_shp['cll']:
            chart_data.add_series(i[0], tuple(i[1:]))
        dct_shp['obj_plt'] = self.lcn['ppt'][ndx_sld]['obj_shp'].add_chart(
            self.dct_chart_type[dct_shp['typ'].lower()],
            self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['left']),
            self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['top']),
            self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['width']),
            self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['height']),
            chart_data,
        ).chart
        dct_shp['obj_plt'].chart_style = dct_shp['sty']  # 图表整体颜色风格[1:48]
        self.gnr_ttl(dct_shp)   # title when dct_shp['ttl']
        self.gnr_axs(dct_shp)   # axis format
        self.gnr_lgd(dct_shp)   # legend
        self.gnr_lbl(dct_shp)   # label


class ppz(tblMixin, pltMixin):
    """
    pptx generator
    >>> lcn = {
    >>>     'fld': '.',
    >>>     'fls': 'tst.pptx',
    >>>     'ppt': {
    >>>         'sld_middle': {
    >>>             'mtr': 'middle',
    >>>             'shp': [
    >>>                 {
    >>>                     'typ': 'textbox',
    >>>                     'tip': '顶栏总结',
    >>>                     'paragraph': [{'ctt': '这里是测试标题', 'fontSize': 16, 'fontBold': True},],
    >>>                 },
    >>>                 {
    >>>                     'typ': 'textbox',
    >>>                     'tip': '左侧内容',
    >>>                     'paragraph': [
    >>>                         {'ctt': '这里是测试的左侧文本内容标题', 'fontSize': 14},
    >>>                         {'ctt': '这里是测试正文，一定要长，测试换行的重要性就在这里体现出来了吗？不知道够不够长了啊，应该不够吧；现在够了吗'},
    >>>                     ],
    >>>                 },
    >>>                 {
    >>>                     'typ': 'table',
    >>>                     'tip': '右侧表格',
    >>>                     'cll': [[12,None,56],[12,34,None],[12,34,56],[{'ctt':12,'color':(225,0,0)},34,56]],
    >>>                     'cll_fit': {
    >>>                         'color': [rgb_rdd,rgb_rdd],
    >>>                         'fontColor': [rgb_wht,rgb_wht],
    >>>                         'fontBold': [True,True],
    >>>                     },
    >>>                     'cll_fit_clm': {'color': [rgb_rdl],},
    >>>                 },
    >>>                 {
    >>>                     'typ': 'picture',
    >>>                     'tip': '右上角logo',
    >>>                     'fld': './pyzohar/samples',
    >>>                     'fls': 'logo_centaline.png',
    >>>                 },
    >>>         ],
    >>>             'tip': '左右平分',
    >>>         },
    >>>         'sld_full': {
    >>>             'mtr': 'full',
    >>>             'shp': [
    >>>                 {
    >>>                     'typ': 'textbox',
    >>>                     'tip': '顶栏总结',
    >>>                     'paragraph': [{'ctt': '这里是测试标题', 'fontSize': 16, 'fontBold': True},],
    >>>                 },
    >>>                 {
    >>>                     'typ':'barchart',
    >>>                     'tip': '主体图表',
    >>>                     'ttl':'演示图表',
    >>>                     'sty': 10,  # [1-48], usually in [9-16], [灰, 彩, 蓝, 红, 绿, 紫, 靛, 橙]
    >>>                     'lbl': 'outside_end',   # 指定数据标签位置, in ['outside_end', 'inside_base']
    >>>                     'lgd': 'rightout',      # 指定图例位置, in ['rightout', 'rightin', 'topout/in', 'bottomout/in']
    >>>                     'clm': ['21-01','21-02','21-03',],  # 横轴分类刻度
    >>>                     'cll': [['ss1',1,2,3],['ss2',1,2,3],],  # 类别和数据
    >>>                 },
    >>>                 {
    >>>                     'typ': 'picture',
    >>>                     'tip': '右上角logo',
    >>>                     'fld': './pyzohar/samples',
    >>>                     'fls': 'logo_centaline.png',
    >>>                 },
    >>>             ],
    >>>             'tip': '全页显示',
    >>>         },
    >>>     },
    >>> }  # sample of mother template 'middle', textbox and table inside
    """
    def __init__(self, dts=None, lcn=None, *, spr=False, cvr=True):
        super(ppz, self).__init__(dts, lcn, spr=spr, cvr=cvr)

    def gnr_pic(self, ndx_sld, dct_shp):
        """
        generate picture
        :param ndx_sld: index of the target slide in self.lcn.ppt
        :param dct_shp: in {'fld','fls','left','top','height','width'}
        """
        w = self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['width']) if 'width' in dct_shp else None
        h = self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['height']) if 'height' in dct_shp else None
        self.lcn['ppt'][ndx_sld]['obj_shp'].add_picture(
            os_join(dct_shp['fld'], dct_shp['fls']),
            self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['left']),
            self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['top']),
            width=w, height=h,
        )

    def gnr_txt(self, ndx_sld, dct_shp):
        """
        generate textbox
        """
        dct_m = self.lcn['ppt']['mtr']
        dct_shp['obj_txt'] = self.lcn['ppt'][ndx_sld]['obj_shp'].add_textbox(
            self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['left']),
            self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['top']),
            self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['width']),
            self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['height']),
        )
        txt = dct_shp['obj_txt'].text_frame
        txt.clear()
        txt.auto_size = MSO_AUTO_SIZE.NONE  # 文本框可以自由拖动，且文字大小与文本框不关联
        txt.vertical_anchor = self.dct_anchor[dct_shp['anchor']] if 'anchor' in dct_shp else self.dct_anchor[
            dct_m['anchor']]  # 纵向对齐（纵向居中）
        txt.word_wrap = True  # 自动换行
        for i_prh in range(len(dct_shp['paragraph'])):
            dct_p = dct_shp['paragraph'][i_prh]
            if i_prh == 0:
                prh = txt.paragraphs[i_prh]
            else:
                prh = txt.add_paragraph()
            prh.line_spacing = self.dct_mtr['line_spacing']  # 行距
            prh.space_before = self.dct_mtr['space_before']  # 段前行距
            prh.space_after = self.dct_mtr['space_after']  # 段后行距
            prh.alignment = self.dct_align[dct_shp['align']] if 'align' in dct_shp else self.dct_align[
                dct_m['align']]  # 文本对齐（两端对齐）
            run = prh.add_run()
            run.text = dct_p['ctt']
            run.font.name = dct_p['fontName'] if 'fontName' in dct_p.keys() else dct_m['fontName']
            fnt_rgb = dct_p['fontColor'] if 'fontColor' in dct_p.keys() else dct_m['fontColor']
            run.font.color.rgb = RGBColor(*fnt_rgb)
            run.font.size = Pt(dct_p['fontSize']) if 'fontSize' in dct_p.keys() else dct_m['fontSize']
            run.font.bold = dct_p['fontBold'] if 'fontBold' in dct_p.keys() else dct_m['fontItalic']
            run.font.italic = dct_p['fontItalic'] if 'fontItalic' in dct_p.keys() else dct_m['fontItalic']

    def gnr_lne(self, ndx_sld, dct_shp):
        """
        generate line
        """
        lns = self.lcn['ppt'][ndx_sld]['obj_shp'].add_shape(
            MSO_SHAPE.RECTANGLE,
            self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['left']),
            self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['top']),
            self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['width']),
            self.dct_util[self.lcn['ppt']['mtr']['utl']](dct_shp['height']),
        )
        lns.line.color.rgb = RGBColor(dct_shp['color'][0], dct_shp['color'][1], dct_shp['color'][2])
        lns.line.color.brightness = 0  # [0,1] => [dark, light]
        lns.line.width = Cm(0.001)
        lns.fill.solid()
        lns.fill.fore_color.rgb = RGBColor(dct_shp['color'][0], dct_shp['color'][1], dct_shp['color'][2])
        lns.shadow.show = False
        lns.shadow.inherit = False

    def gnr_shp(self, ndx_sld):
        """
        generate shapes
        """
        self.lcn['ppt'][ndx_sld]['obj_shp'] = self.lcn['ppt'][ndx_sld]['obj_sld'].shapes
        for dct_k in self.lcn['ppt'][ndx_sld]['shp']:  # 遍历dict of shapes
            if dct_k['typ'] in ['textbox']:
                self.gnr_txt(ndx_sld, dct_k)
            elif dct_k['typ'] in ['line']:
                self.gnr_lne(ndx_sld, dct_k)
            elif dct_k['typ'] in ['picture']:
                self.gnr_pic(ndx_sld, dct_k)
            elif dct_k['typ'] in ['table']:
                self.gnr_tbl(ndx_sld, dct_k)
            elif dct_k['typ'].lower() in ['plot', 'chart', 'barchart', 'linechart', 'piechart']:
                self.gnr_plt(ndx_sld, dct_k)

    def gnr_sld(self):
        """
        generate slides
        """
        for ndx_sld in [i for i in self.lcn['ppt'] if i != 'mtr']:  # ['sld_1']
            print('info: generate slide %s' % self.lcn['ppt'][ndx_sld]['tip'])
            self.sld_mtr(ndx_sld)  # 自定义布局母版检查
            sld_lyt = self.ppt.slide_layouts[6]
            self.lcn['ppt'][ndx_sld]['obj_sld'] = self.ppt.slides.add_slide(sld_lyt)
            self.gnr_shp(ndx_sld)

    def ppt_xpt(self):
        """"""
        self.ppt.save(os_join(self.lcn['fld'], self.lcn['fls']))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 22 16:30:38 2019
basic type's alteration.
@author: zoharslong
"""
from datetime import datetime as dt_datetime
from datetime import date as typ_dt_date, time as typ_dt_time, timedelta as typ_dt_timedelta
from time import struct_time as typ_tm_structtime, strftime as tm_strftime, mktime as tm_mktime
from bson.objectid import ObjectId as typ_ObjectId  # _id from mongobd
from calendar import monthrange     # how many days in any month
from re import search as re_search, sub as re_sub, match as re_match
from math import isnan as math_isnan
from collections.abc import Iterable
from random import randint
from hashlib import md5 as hsh_md5
from numpy import array as np_array, ndarray as typ_np_ndarray
from pandas.core.series import Series as typ_pd_Series                  # 定义series类型
from pandas.core.frame import DataFrame as typ_pd_DataFrame             # 定义dataframe类型
from pandas.core.indexes.base import Index as typ_pd_Index              # 定义dataframe.columns类型
from pandas.core.indexes.range import RangeIndex as typ_pd_RangeIndex   # 定义dataframe.index类型
from pandas._libs.tslibs.nattype import NaTType as typ_pd_NaT
from pandas._libs.tslib import Timestamp as typ_pd_Timestamp
from pandas import DataFrame as pd_DataFrame
from pandas import to_datetime as pd_to_datetime
dct_ncd_int = {
    '0': ['IJ', 'KL', 'MN', 'OP', 'QR'],
    '1': ['GH', 'IJ', 'KL', 'MN', 'OP'],
    '2': ['EF', 'GH', 'IJ', 'KL', 'MN'],
    '3': ['CD', 'EF', 'GH', 'IJ', 'KL'],
    '4': ['AB', 'CD', 'EF', 'GH', 'IJ'],
    '5': ['ZY', 'XW', 'VU', 'TS', 'AB'],
    '6': ['XW', 'VU', 'TS', 'RQ', 'CD'],
    '7': ['VU', 'TS', 'RQ', 'AB', 'EF'],
    '8': ['TS', 'RQ', 'PO', 'CD', 'GH'],
    '9': ['RQ', 'PO', 'AB', 'EF', 'ZY'],
}   # 一个简单的君子加密
dct_tms_fTr = {
    '%Y-%m-%d %H:%M:%S ': '^[0-9]{4}-[0-9]{1,2}-[0-9]{1,2} [0-9]{1,2}:[0-9]{1,2}:[0-9]{1,2} $',
    '%Y-%m-%d %H:%M:%S': '^[0-9]{4}-[0-9]{1,2}-[0-9]{1,2} [0-9]{1,2}:[0-9]{1,2}:[0-9]{1,2}$',
    '%Y-%m-%d %H:%M:%S.%f': '^[0-9]{4}-[0-9]{1,2}-[0-9]{1,2} [0-9]{1,2}:[0-9]{1,2}:[0-9]{1,2}[.][0-9]+',
    '%Y-%m-%d': '^[0-9]{4}-[0-9]{1,2}-[0-9]{1,2}$',
    '%Y/%m/%d %H:%M:%S': '^[0-9]{4}/[0-9]{1,2}/[0-9]{1,2} [0-9]{1,2}:[0-9]{1,2}:[0-9]{1,2}$',
    '%Y/%m/%d': '^[0-9]{4}/[0-9]{1,2}/[0-9]{1,2}$',
    '%Y.%m.%d %H:%M:%S': '^[0-9]{4}[.][0-9]{1,2}[.][0-9]{1,2} [0-9]{1,2}:[0-9]{1,2}:[0-9]{1,2}$',
    '%Y.%m.%d': '^[0-9]{4}[.][0-9]{1,2}[.][0-9]{1,2}$',
    '%Y年%m月%d日 %H:%M:%S': '^[0-9]{2,4}年[0-9]{1,2}月[0-9]{1,2}日 [0-9]{1,2}:[0-9]{1,2}:[0-9]{1,2}$',
    '%Y年%m月%d日 %H时%M分%S秒': '^[0-9]{2,4}年[0-9]{1,2}月[0-9]{1,2}日 [0-9]{1,2}时[0-9]{1,2}分[0-9]{1,2}秒$',
    '%Y年%m月%d日': '^[0-9]{2,4}年[0-9]{1,2}月[0-9]{1,2}日$',
    'int': '^[0-9]+[.]{0}[0-9]*$',
    'float': '^[0-9]+[.]{1}[0-9]*$',
    '%Yw%w': '^[0-9]{4}[w][0-9]{2}$',
    '%Ym%m': '^[0-9]{4}[m][0-9]{2}$',
    '%yw%w': '^[0-9]{2}[w][0-9]{2}$',
    '%ym%m': '^[0-9]{2}[m][0-9]{2}$',
}   # 一组常用的时间类format/regex对应字典


class stz(str):
    """
    type str altered by zoharslong.
    >>> stz('1989-02-14').add_rgx(rtn=True)     # generate regex of the string content
    ['^[0-9]{4}-[0-9]{1,2}-[0-9]{1,2}$']
    >>> stz('1989-02-14').add_fmt(rtn=True)     # generate format of the string content
    ['%Y-%m-%d']
    >>> stz('15814079977').ncd_int(rtn=True)
    'IJXWRQIJCDKLTSPOPOTSTS1'
    >>> stz('IJXWRQIJCDKLTSPOPOTSTS1').dcd_int(rtn=True)
    '15814079977'
    >>> stz('15814079977').ncd_md5(rtn=True)
    'f8ceb85867e2df236e11211366be6479'
    >>> stz('15814079977').ncd_cvr([3,7], rtn=True)
    '158****9977'
    """
    def __init__(self, val=''):
        """
        create a new string object from the given object.
        :param val: 指定的值
        """
        super().__init__()
        self.__val, self.rgx, self.fmt, self.cod = val, None, None, None

    def add_rgx(self, *, lst_rgx=None, rtn=False):
        """
        match regex.
        从传入的正则列表中筛选出符合类的正则表达式组成列表.
        :param lst_rgx: 参与匹配的正则表达式列表
        :param rtn: return or not, default False
        :return: None
        """
        lst_rgx = list(dct_tms_fTr.values()) if lst_rgx is None else lst_rgx
        rgx_xpt = [i_rgx for i_rgx in lst_rgx if re_search(i_rgx, self)]
        self.rgx = rgx_xpt
        if rtn:
            return self.rgx

    def add_fmt(self, *, dct_rgx=None, rtn=False):
        """
        match format.
        从传入的格式-正则字典中筛选出符合类的格式组成列表.
        :param dct_rgx: 参与匹配的格式-正则表达式字典
        :param rtn: return or not, default False
        :return: None
        """
        if self.rgx is None or len(self.rgx) <= 0:
            self.add_rgx()
        self.fmt = [] if self.fmt is None else self.fmt
        dct_rgx = dct_tms_fTr if dct_rgx is None else dct_rgx
        for i in range(len(self.rgx)):
            self.fmt.append(list(dct_rgx.keys())[list(dct_rgx.values()).index(self.rgx[i])])
        if rtn:
            return self.fmt

    def ncd_int(self, *, dct_ncd=None, rtn=False):
        """
        数字类字符串加密，特别用于电话号码.
        :param dct_ncd: 默认转换字典dct_ncd_int
        :param rtn: if rtn, return the target self.cod, default False
        :return: return self.cod
        """
        try:
            if re_match("^\d+?$", self) is not None:
                self.cod = self
                dct_ncd = dct_ncd_int if dct_ncd is None else dct_ncd
                int_vrt = randint(0, len(dct_ncd['0']) - 1)
                lst_vrt = [dct_ncd[str(i)][int_vrt] for i in range(0, 10)]
                for i in range(0, 10):
                    self.cod = re_sub(str(i), lst_vrt[i], self.cod)
                self.cod += str(int_vrt)
            if rtn:
                return self.cod
        except TypeError:
            if rtn:
                return self

    def dcd_int(self, *, dct_dcd=None, rtn=False):
        """
        数字类字符串解密.
        :param dct_dcd: 默认转换字典dct_ncd_int
        :param rtn: if rtn, return the target self.cod, default False
        :return: return self.cod
        """
        try:
            if re_match("^[A-Z]+[0-9]+$", self) is not None:
                self.cod = self
                dct_ncd = dct_ncd_int if dct_dcd is None else dct_dcd
                int_vrt = int(re_sub('[A-Z]', '', self.cod))
                lst_vrt = [dct_ncd[str(i)][int_vrt] for i in range(0, 10)]
                self.cod = re_sub('[0-9]', '', self.cod)
                for i in range(0, 10):
                    self.cod = re_sub(lst_vrt[i], str(i), self.cod)
            if rtn:
                return self.cod
        except TypeError:
            if rtn:
                return self

    def ncd_md5(self, *, prm=None, rtn=False):
        """
        使用哈希MD5对文本进行加密后转化为唯一标记.
        :param prm: the encoding code
        :param rtn: if rtn is True, return self.cod
        :return: a str of code
        """
        prm = 'zohar' if not prm else prm   # 默认的哈希参数
        self.cod = self
        if type(self) in [str, stz]:
            hsh = hsh_md5(bytes(prm, encoding='utf-8'))
            hsh.update(bytes(self.cod, encoding="utf-8"))
            self.cod = hsh.hexdigest()
        if rtn:
            return self.cod

    def ncd_cvr(self, lst_cvr, *, rtn=True):
        """
        encode by covering '*'
        :param lst_cvr: cover targeted index by *
        :param rtn: if rtn is True, return self.cod
        :return: a str of code
        """
        bgn = lst_cvr[0]
        end = len(self) + 1 + lst_cvr[1] if lst_cvr[1] < 0 else lst_cvr[1]
        lst = list(self)
        for i in range(bgn, end):
            if len(lst) >= i:
                lst[i] = '*'
        self.cod = ''.join(lst)
        if rtn:
            return self.cod


class lsz(list):
    """
    type list altered by zoharslong.
    >>> lsz([[1,2,3],[1,2]]).edg_of_len(rtn=True)
    [2, 3]
    >>> lsz({'A':1,'B':2,'C':3}).typ_to_lst(rtn=True)
    [('A', 1), ('B', 2), ('C', 3)]
    >>> lsz([1,2,3]).lst_to_typ('dict', ['A','B','C'], rtn=True)
    [{'A': 1, 'B': 2, 'C': 3}]
    >>> lsz([1,2,3]).cpy_tal(5, rtn=True)
    [1, 2, 3, 3, 3]
    >>> lsz([1,2,3]).mrg_cll(['_A','_B','_C'], rtn=True)
    ['1_A', '2_B', '3_C']
    >>> lsz([[1,2],3,[4,5,[6,7,[8,9]]]]).nfd(rtn=True)
    [1, 2, 3, 4, 5, 6, 7, 8, 9]
    >>> lsz().mrg('inter', [1, 2, 3], [2, 3, 4], rtn=True)
    [2, 3]
    >>> lsz(['O']).chk_dtf_dtp(pd_DataFrame([{"A":1,"B":"a"},{"A":2,"B":"b"}]), rtn=True)
    ['B']
    """
    __slots__ = ('__seq', 'typ', 'len', 'len_min', 'len_max')
    lst_typ_lsz = [
        str,
        stz,
        int,
        float,
        list,
        dict,
        tuple,
        typ_np_ndarray,
        typ_pd_DataFrame,
        typ_pd_Series,
        typ_pd_Index,
        typ_pd_RangeIndex,
        type(None)
    ]   # lsz.seq's type

    def __init__(self, seq=None, spr=True, lst=False):
        """
        create a new list object from the given object.
        :param seq: first, save target into lsz.seq
        :param spr: let lsz = lsz.seq or not, default False
        :param lst:
        """
        super().__init__()
        self.__seq, self.typ, self.len, self.len_max, self.len_min = None, None, None, None, None
        self.__init_rst(seq, spr, lst)

    def __init_rst(self, seq=None, spr=False, lst=False):
        """
        private reset initiation.
        :param seq: a list content in any type, None for []
        :param spr: let lsz = lsz.seq or not, default False
        :return: None
        """
        if self.seq is None:
            self.seq = seq if seq is not None else []
        if lst:
            self.typ_to_lst()
        if spr:
            self.spr_nit()

    def spr_nit(self, rtn=False, prn=False):
        """
        super initiation. let lsz = lsz.seq.
        :param rtn: return lsz or not, default False
        :param prn: if print or not, default False
        :return: if rtn is True, return lsz
        """
        try:
            super(lsz, self).__init__(self.__seq)
        except TypeError:
            if prn:
                print('info: %s cannot convert to list.' % (str(self.__seq)[:8]+'..'))
        if rtn:
            return self

    @property
    def seq(self):
        """
        @property get & set lsz.seq.
        :return: lsz.seq
        """
        return self.__seq

    @seq.setter
    def seq(self, seq):
        """
        lsz.seq = seq.
        :param seq: a sequence to import.
        :return: None
        """
        if type(seq) in self.lst_typ_lsz:
            self.__seq = seq
            self.__attr_rst()
        else:
            raise TypeError('seq\'s type %s is not available.' % type(seq))

    def edg_of_len(self, rtn=False):
        """
        get the max and min length of cells in lsz.seq.
        :param rtn: return the result or not, default False
        :return: if rtn is True, return the result in format [min, max]
        """
        lst_len = [len(i) for i in self.seq if isinstance(i, Iterable)]
        self.len_max = max(lst_len) if lst_len != [] else None
        self.len_min = min(lst_len) if lst_len != [] else None
        if rtn:
            return [self.len_min, self.len_max]

    def __attr_rst(self, prn=False):
        """
        reset attributes lsz.typ.
        :return: None
        """
        self.typ = type(self.__seq)
        try:
            self.len = len(self.__seq)
            self.edg_of_len()
        except TypeError:
            if prn:
                print('info: %s is not available for __len__().' % (str(self.__seq)[:8] + '..'))

    def typ_to_lst(self, *, spr=False, rtn=False, prm='records'):
        """
        alter lsz.seq's type to list.
        :param spr: let lsz = lsz.seq or not, default False
        :param rtn: return the result or not, default False
        :param prm: if lsz.seq is pd.DataFrame, method in ['dict', 'list', 'series', 'split', 'records', 'index']
        :return: if rtn is True, return the final result
        """
        if self.typ in [list]:
            pass
        elif self.typ in [dict]:
            self.seq = list(self.seq.items())
        elif self.typ in [tuple, typ_pd_Series, typ_pd_Index, typ_pd_RangeIndex]:
            self.seq = list(self.seq)
        elif self.typ in [typ_np_ndarray]:
            self.seq = self.seq.tolist()
        elif self.typ in [typ_pd_DataFrame]:
            self.seq = self.seq.to_dict(orient=prm)
        else:
            self.seq = [self.seq]
        if spr:
            self.spr_nit()
        if rtn:
            return self.seq

    def lst_to_typ(self, str_typ='list', *args, spr=False, rtn=False):
        """
        alter list to other type.
        :param str_typ: target data type
        :param args: if needs other list to merge, import here, len(args) in [1,2]
        :param spr: let lsz = lsz.seq or not, default False
        :param rtn: return the result or not, default False
        :return: if rtn is True, return the final result
        """
        self.typ_to_lst()
        if str_typ.lower() in ['list', 'lst']:
            pass
        elif str_typ.lower() in ['listt', 'lstt', 't']:
            self.seq = np_array(self.seq).T.tolist()
        elif str_typ.lower() in ['str']:  # 用于将list转化为不带引号的字符串:[1,'a',...] -> "1,a,..."
            self.seq = str(self.seq).replace("'", '')[1:-1]
        elif str_typ.lower() in ['dict', 'dct', '[dict]', '[dct]', 'listdict', 'lstdct']:  # [a,b]+[1,1]->[{a:1},{b:1}]
            self.seq = args[-1] if not self.seq else self.seq  # 当lsz.seq为空时，从*args中依次取值
            self.seq = pd_DataFrame([self.seq], columns=lsz(args[0]).typ_to_lst(rtn=True)).to_dict(orient='records')
        elif str_typ.lower() in ['listtuple', 'lsttpl', 'list_tuple', 'lst_tpl']:   # [a,b,..]+[1,1,..] ->[(a,1),..]
            if not self.seq:  # 当lsz.seq为空时，从*args中依次取值
                int_bgn, self.seq = -1, args[0]
            else:
                int_bgn = 0
            lst_prm = lsz(args[int_bgn]).cpy_tal(self.len, rtn=True)
            self.seq = [(i, lst_prm[self.seq.index(i)]) for i in self.seq]
        if spr:
            self.spr_nit()
        if rtn:
            return self.seq

    def cpy_tal(self, flt_len, *, spr=False, rtn=False):
        """
        copy the last cell to tails until lsz.len equals to flt_len.
        :param flt_len: the target length of lsz after running this def
        :param spr: let lsz = lsz.seq or not, default False
        :param rtn: return the result or not, default False
        :return: if rtn is True, return the final result
        """
        self.typ_to_lst()
        lst_xpt = self.seq.copy()
        while len(lst_xpt) < flt_len:
            lst_xpt.append(self.seq[-1])
        self.seq = lst_xpt
        if spr:
            self.spr_nit()
        if rtn:
            return self.seq

    def mrg_cll(self, *args, spr=False, rtn=False):
        """
        in lsz.seq, cell to cell merging.
        structuring: ['a','b',...] + ['x'] -> ['ax','bx',...].
        :param args: lists to merge in method cell by cell
        :param spr: let lsz = lsz.seq or not, default False
        :param rtn: return the result or not, default False
        :return: if rtn is True, return the final result
        """
        self.typ_to_lst()
        flt_max = lsz(args).edg_of_len(True)[1]
        if not self.seq:    # 当lsz.seq为空时，从*args中依次取值
            int_bgn, self.seq = 1, args[0]
        else:               # 当lsz.seq有值时，该list也加入计算
            int_bgn = 0
        self.cpy_tal(flt_max)
        for i in range(int_bgn, len(args)):
            lst_mrg = lsz(args[i]).cpy_tal(flt_max, rtn=True)
            self.seq = [str(self.seq[i]) + str(lst_mrg[i]) for i in range(self.len)]
        if spr:
            self.spr_nit()
        if rtn:
            return self.seq

    def nfd(self, spr=False, rtn=False):
        """
        unfold lists in lsz.seq.
        >>> lsz([[1,2],3,[4,5,[6,7,[8,9]]]]).nfd(rtn=True)
        [1, 2, 3, 4, 5, 6, 7, 8, 9]
        :param spr: let lsz = lsz.seq or not, default False
        :param rtn: return the result or not, default False
        :return: if rtn is True, return the final result
        """
        self.typ_to_lst()
        while [True for i_cll in self.seq if type(i_cll) in [list, tuple]]:     # check if there is any cell in list
            lst_nfd = []
            for i_cll in self.seq:
                if type(i_cll) in [list]:
                    for i in i_cll:
                        lst_nfd.append(i)
                else:
                    lst_nfd.append(i_cll)
            self.seq = lst_nfd
        if spr:
            self.spr_nit()
        if rtn:
            return self.seq

    def mrg(self, str_mtd, *args, spr=False, rtn=False):
        """
        merge lists in method intersection, difference or union.
        >>> lsz().mrg('inter', False, True, [1, 2, 3], [2, 3, 4])
        [2, 3]
        :param str_mtd: ['inter','differ','union']
        :param args: lists to be merged
        :param spr: let lsz = lsz.seq or not, default False
        :param rtn: return the result or not, default False
        :return: if rtn is True, return the final result
        """
        self.typ_to_lst()
        if not self.seq:    # 当lsz.seq为空时，从*args中依次取值
            int_bgn, self.seq = 1, args[0]
        else:               # 当lsz.seq有值时，该list也加入计算
            int_bgn = 0
        for i in range(int_bgn, len(args)):
            if str_mtd.lower() in ['intersection', 'inter']:    # 交
                self.seq = list(set(self.seq) & set(args[i]))
            elif str_mtd.lower() in ['difference', 'differ']:   # 差
                self.seq = list(set(self.seq) ^ set(args[i]))
            elif str_mtd.lower() in ['union']:                  # 并
                self.seq = list(set(self.seq) | set(args[i]))
        if spr:
            self.spr_nit()
        if rtn:
            return self.seq

    def chk_dtf_dtp(self, dtf_chk, *, spr=False, rtn=False, prm=None):
        """
        check dataframe's columns' dtype.
        >>> lsz(['O']).chk_dtf_dtp(pd_DataFrame([{"A":1,"B":"a"},{"A":2,"B":"b"}]),rtn=True)
        ['B']
        >>> lsz().chk_dtf_dtp(pd_DataFrame([{"A":1,"B":"a"},{"A":2,"B":"b"}]), prm='O',rtn=True)
        ['B']
        :param dtf_chk: the target dataframe to be listing
        :param spr: let lsz = lsz.seq or not, default False
        :param rtn: return the result or not, default False
        :param prm: a list of columns' dtype, such as ['O','int64','float64',...], can be shown by dataframe.dtypes
        :return: if rtn is True, return the final result
        """
        self.typ_to_lst()
        lst_clm_dtp = []
        if self.seq is None and prm is not None:
            self.seq = lsz(prm).typ_to_lst(rtn=True)
        for i_dtp in self.seq:
            lst_clm_dtp.extend(dtf_chk.dtypes[dtf_chk.dtypes == i_dtp].index.values)
        self.seq = lst_clm_dtp
        if spr:
            self.spr_nit()
        if rtn:
            return self.seq

    def rgs_to_typ(self, *, spr=False, rtn=False, prm=None, prn=False):
        """
        args in functions to list. 仅用于函数中对*args的处理.
        >>> lsz(('a',)).rgs_to_typ(rtn=True)
        ['a', 'a']
        >>> lsz(({'a':'b'},)).rgs_to_typ(rtn=True)
        [['a'], ['b']]
        >>> lsz((['a','b'],'x')).rgs_to_typ(prm='dct', rtn=True)
        {'a': 'x', 'b': 'x'}
        >>> lsz((['a','c'],'b')).rgs_to_typ(prm='equal',rtn=True)
        [['a', 'c'], ['b', 'b']]
        >>> lsz(('a',['b','c'])).rgs_to_typ(prm='eql',rtn=True)
        [['a', 'a'], ['b', 'c']]
        :param prm: in ['equal', 'eql', 'dict', 'dct']
        :param spr: super self.seq to self or not, default False
        :param rtn: return or not, default False
        :param prn: print or not, default False
        :return: None if not rtn
        """
        if self.seq in [None, []]:
            pass
        elif type(self.seq[0]) is dict and self.len == 1:   # args from ({a:A,b:B},) to [[a, b], [A, B]]
            str_mpt = list(self.seq[0].keys())
            str_xpt = list(self.seq[0].values())
            self.seq = [str_mpt, str_xpt]
        elif self.len == 1:                                 # args from (x,) to [x, x]
            self.typ_to_lst()
            self.cpy_tal(2)
        else:                                               # args from (x, y, ...) to [x, y, ...]
            self.typ_to_lst()
        if prm in ['dct', 'dict', 'dictionary'] and self.seq not in [None, []]:  # args from [[x,y],[a,b]] to {x:y,a:b}
            arg = self.seq.copy()
            self.seq = arg[0]
            lst_bgn = self.typ_to_lst(rtn=True)
            self.seq = arg[1]
            self.cpy_tal(len(lst_bgn))
            self.seq = self.lst_to_typ('dct', lst_bgn, rtn=True)[0]
        elif prm in ['eql', 'equal']:                       # args from [x, [a,b]] to [[x,x],[a,b]]
            arg = self.seq.copy()
            self.seq = arg[0]
            lst_bgn = self.typ_to_lst(rtn=True)
            self.seq = arg[1]
            lst_end = self.typ_to_lst(rtn=True)
            self.seq = [lst_bgn, lst_end]
            len_fnl = self.len_max
            if len(lst_bgn) < len_fnl:
                self.seq = lst_bgn.copy()
                self.cpy_tal(len_fnl)
                lst_bgn = self.seq
            elif len(lst_end) < len_fnl:
                self.seq = lst_end.copy()
                self.cpy_tal(len_fnl)
                lst_end = self.seq
            self.seq = [lst_bgn, lst_end]
        elif prn:
            print('info: prm needs ["dct","eql"] for format [{x:y}, [[x,x],[a,b]]].')
        else:
            pass
        if spr:
            self.spr_nit()
        if rtn:
            return self.seq


class dcz(dict):
    """
    type dict altered by zoharslong.
    >>> dcz().typ_to_dct(['A','B','C'], [1,0], rtn=True)
    {'A': 1, 'B': 0, 'C': 0}
    >>> dcz([1,0]).typ_to_dct(['A','B','C'], rtn=True)
    {'A': 1, 'B': 0, 'C': 0}
    >>> dcz([{'A':1},{'B':2}]).nfd(rtn=True)
    {'A': 1, 'B': 2}
    >>> dcz({'A':{'B':1,'C':2},'D':3}).nfd(rtn=True)
    {'B': 1, 'C': 2, 'D': 3}
    """
    __slots__ = ('__seq', 'typ', 'len', 'kys', 'vls')
    lst_typ_dcz = [
        str,
        stz,
        int,
        float,
        list,
        dict,
        tuple,
        typ_np_ndarray,
        typ_pd_DataFrame,
        typ_pd_Series,
        typ_pd_Index,
        typ_pd_RangeIndex,
        type(None)
    ]   # lsz.seq's type

    def __init__(self, seq=None, spr=True):
        """
        create a new list object from the given object.
        :param seq: first, save target into dcz.seq
        :param spr: let dcz = dcz.seq or not, default False
        """
        super().__init__()
        self.__seq, self.typ, self.len, self.kys, self.vls = None, None, None, None, None
        self.__init_rst(seq, spr)

    def __init_rst(self, seq=None, spr=False):
        """
        private reset initiation.
        :param seq: a list content in any type, None for []
        :param spr: let dcz = dcz.seq or not, default False
        :return: None
        """
        if self.seq is None:
            self.seq = seq if seq is not None else {}
        if spr:
            self.spr_nit()

    def spr_nit(self, rtn=False):
        """
        super initiation. let dcz = dcz.seq.
        :param rtn: return dcz or not, default False
        :return: if rtn is True, return dcz
        """
        try:
            self.clear()    # 由于直接super会导致dcz.seq拼接到原dcz之后，因此先调用dcz.clear()清空原dcz
            super(dcz, self).__init__(self.__seq)
        except (TypeError, ValueError):
            print('info: %s cannot convert to list.' % (str(self.__seq)[:8]+'..'))
        if rtn:
            return self

    @property
    def seq(self):
        """
        @property get & set lsz.seq.
        :return: dcz.seq
        """
        return self.__seq

    @seq.setter
    def seq(self, seq):
        """
        dcz.seq = seq.
        :param seq: a sequence to import.
        :return: None
        """
        if type(seq) in self.lst_typ_dcz:
            self.__seq = seq
            self.__attr_rst()
        else:
            raise TypeError('seq\'s type %s is not available.' % type(seq))

    def __attr_rst(self, prn=False):
        """
        reset attributes dcz.typ.
        :return: None
        """
        self.typ = type(self.__seq)
        try:
            self.len = len(self.__seq)
        except TypeError:
            if prn:
                print('info: %s is not available for __len__().' % (str(self.__seq)[:8] + '..'))
        self.kys = list(self.__seq.keys()) if self.typ in [dict] else None
        self.vls = list(self.__seq.values()) if self.typ in [dict] else None

    def typ_to_dct(self, *args, spr=False, rtn=False):
        """
        alter dcz.seq's type from others to dict.
        :param args: if needs other objects to merge, import here, len(args) in [1,2]
        :param spr: let dcz = dcz.seq or not, default False
        :param rtn: return the result or not, default False
        :return: if rtn is True, return the final result
        """
        if self.seq and self.typ in [dict]:
            pass
        elif (self.seq and self.typ in [list, lsz]) or (not self.seq and [True for i in args if type(i) in [list]]):
            lst_clm = lsz(args[0]).typ_to_lst(rtn=True)
            # 当dcz.seq为空时，从*args中依次取值
            self.seq = args[-1] if not self.seq else self.seq           # dcz.seq为空则取末尾的args
            self.seq = lsz(self.seq).cpy_tal(len(args[0]), rtn=True)    # dcz.seq填充至长度等同于args[0]即dict的keys
            self.seq = pd_DataFrame([self.seq], columns=lst_clm).to_dict(orient='records')[0]
        if spr:
            self.spr_nit()
        if rtn:
            return self.seq

    def nfd(self, spr=False, rtn=False):
        """
        unfold dict cells inside a list or dict object.
        :param spr: let dcz = dcz.seq or not, default False
        :param rtn: return the result or not, default False
        :return: if rtn is True, return the final result in dict format {A,B,..}
        """
        dct_tmp = {}
        if self.typ in [list, lsz]:     # [{A:1},{B:2},...] -> {A:1,B:2}
            for i_dct in self.seq:
                if type(i_dct) is dict:
                    dct_tmp.update(i_dct)
        elif self.typ in [dict]:        # {A:{B:1,C:2},D:3,...} -> {B:1,C:2,D:3,...}
            for i_dct in self.seq:
                if type(self.seq[i_dct]) is dict:
                    dct_tmp.update(self.seq[i_dct])
                else:
                    dct_tmp.update({i_dct: self.seq[i_dct]})
        self.seq = dct_tmp
        if spr:
            self.spr_nit()
        if rtn:
            return self.seq


class dtz(object):
    """
    type datetime altered by zoharslong.
    >>> dtz(15212421.1).typ_to_dtt(rtn=True)    # from any type to datetime
    datetime.datetime(1970, 6, 26, 9, 40, 21)
    >>> dtz(dtz(15212421.1).typ_to_dtt(rtn=True)).dtt_to_typ(rtn=True)   # from datetime to any type
    '1970-06-26'
    >>> dtz(dtz(15212421.1).typ_to_dtt(rtn=True)).shf(5, rtn=True)    # 5 days after 1970-06-26 in type datetime
    datetime.datetime(1970, 7, 1, 9, 40, 21)
    >>> dtz('2019m03').prd_to_dtt(-1, rtn=True) # get the last day in 2019 March
    datetime.datetime(2019, 3, 31, 0, 0)
    >>> dtz('2019w03').prd_to_dtt(-1, rtn=True) # get the last day in 2019 3th week
    datetime.datetime(2019, 3, 31, 0, 0)
    """
    __slots__ = ('__val', 'typ', 'len', 'fmt',)
    lst_typ_dtz = [
        type,
        str,
        stz,
        int,
        float,
        list,
        typ_pd_NaT,
        dt_datetime,
        typ_dt_date,
        typ_dt_time,
        typ_pd_Timestamp,
        typ_tm_structtime,
        typ_ObjectId,
        type(None)
    ]   # dtz.val's type

    def __init__(self, val=None, prm=None):
        """
        initiating dtz.val, dtz.typ.
        :param val: a datetime content in any type, None for datetime.datetime.now()
        :param prm: 默认None不指定当初始化为空时返回当前时间，任意赋值（如'now'）则赋值当前时间
        """
        self.__val, self.typ, self.len, self.fmt = None, None, None, None
        self.__init_rst(val, prm)

    def __init_rst(self, val=None, prm=None):
        """
        private reset initiation.
        :param val: a datetime content in any type, None for datetime.datetime.now()
        :param prm: 默认None不指定当初始化为空时返回当前时间，任意赋值（如'now'）则赋值当前时间
        :return: None
        """
        if val == 'now' or (self.val is None and prm is not None):
            self.val = dt_datetime.now()
        else:
            self.val = val

    @property
    def val(self):
        """
        @property get & set dtz.val.
        :return: dtz.val
        """
        return self.__val

    @val.setter
    def val(self, val):
        """
        dtz.val = val.
        :param val: a value to import.
        :return: None
        """
        if type(val) in self.lst_typ_dtz:
            self.__val = val
            self.__attr_rst()
        else:
            raise TypeError('info: val\'s type %s is not available.' % type(val))

    def __attr_rst(self):
        """
        reset attributes dtz.typ, dtz.fmt.
        :return: None
        """
        self.typ = type(self.__val)
        self.len = len(self.__val) if self.typ in [str, stz] else None
        self.fmt = stz(self.__val).add_fmt(rtn=True) if self.typ == str else None

    def __str__(self):
        """
        print(className).
        :return: None
        """
        if type(self.__val) == typ_tm_structtime:
            str_xpt = 'time.struct_time'+str(tuple(self.__val))
        else:
            str_xpt = self.__val
        return '%s' % str_xpt
    __repr__ = __str__  # 调用类名的输出与print(className)相同

    def typ_to_dtt(self, rtn=False):
        """
        alter other type to datetime.datetime type.
        fit for datetime.datetime, pd.Timestamp, tm.structtime, float, int and str.
        :param rtn: return the result or not, default False
        :return: if rtn is True, return dtz.val
        """
        if self.typ in [dt_datetime]:
            pass
        elif self.typ in [typ_pd_NaT, type(None)]:
            self.val = None
        elif self.typ in [typ_dt_date]:
            self.val = dt_datetime.combine(self.val, typ_dt_time())
        elif self.typ in [typ_pd_Timestamp]:
            self.val = dt_datetime.combine(self.val.date(), self.val.time())
        elif self.typ in [typ_tm_structtime]:
            self.val = dt_datetime.strptime(tm_strftime('%Y-%m-%d %H:%M:%S', self.val), '%Y-%m-%d %H:%M:%S')
        elif self.typ in [typ_ObjectId]:
            self.val = tm_strftime('%Y-%m-%d %H:%M:%S', self.val.generation_time.timetuple())
            self.val = dt_datetime.strptime(self.val, '%Y-%m-%d %H:%M:%S')
        elif self.typ in [int, float]:
            if not math_isnan(self.val):    # 正常情况：非空float或int
                self.val = dt_datetime.fromtimestamp(int(str(self.val).rsplit('.')[0]))
            else:                           # 对np.nan赋空
                self.val = None
        elif self.typ in [str, stz]:
            if [True for i in self.fmt if i in ['float', 'int']]:
                self.val = dt_datetime.fromtimestamp(int(self.val.rsplit('.')[0]))
            else:
                try:                # 正常情况：匹配正则表达式后转为datatime
                    if self.fmt[0] in ['%yw%w', '%Yw%w']:
                        self.prd_to_dtt()
                    else:
                        self.val = dt_datetime.strptime(self.val, self.fmt[0])
                except IndexError:  # 对无法找到合适正则匹配（如''空字符串）则赋空
                    self.val = None
                    print('info: %s cannot convert to datetime type.' % str(self.val))
        else:
            raise TypeError('type of value not in [dt_datetime, dt_date, tm_structtime, pd_Timestamp, int, float, str]')
        if rtn:
            return self.val

    def dtt_to_typ(self, str_typ='str', str_fmt='%Y-%m-%d', *, rtn=False):
        """
        alter datetime.datetime type to other type.
        fit for datetime.datetime, pd.Timestamp, tm.structtime, float, int and str.
        :param str_typ: target type
        :param str_fmt: if str_typ in ['str'],
        :param rtn: return the result or not, default False
        :return: if rtn is True, return self.val in type str_typ, format str_fmt
        """
        if self.typ not in [dt_datetime]:
            self.typ_to_dtt()
        if str_typ.lower() in ['datetime', 'dt_datetime', 'dt'] or self.val is None:
            pass
        elif str_typ.lower() in ['date', 'dt_date']:
            self.val = self.val.date()
        elif str_typ.lower() in ['time', 'dt_time']:
            self.val = self.val.time()
        elif str_typ.lower() in ['hour']:
            self.val = self.val.hour()
        elif str_typ.lower() in ['week']:
            self.val = self.val.isocalendar()[1]
        elif str_typ.lower() in ['weekDay', 'weekday']:
            self.val = self.val.isocalendar()[2]
        elif str_typ.lower() in ['day']:
            self.val = self.val.day
        elif str_typ.lower() in ['month', 'mnth']:
            self.val = self.val.month
        elif str_typ.lower() in ['year']:
            self.val = self.val.year
        elif str_typ.lower() in ['pd_structtime', 'pd_timestamp']:
            self.val = pd_to_datetime(dt_datetime.strftime(self.val, '%Y-%m-%d %H:%M:%S'), format='%Y-%m-%d %H:%M:%S')
        elif str_typ.lower() in ['timetuple', 'structtime', 'tm_structtime']:
            self.val = self.val.timetuple()
        elif str_typ.lower() in ['mday']:
            self.val = self.val.timetuple().tm_mday
        elif str_typ.lower() in ['yday']:
            self.val = self.val.timetuple().tm_yday
        elif str_typ.lower() in ['float', 'flt', 'int']:
            self.val = int(tm_mktime(self.val.timetuple()))
        elif str_typ.lower() in ['string', 'str', 'stz']:
            if str_fmt in ['%Y年%m月%d日']:
                self.val = dt_datetime.strftime(self.val, '%Y{y}%m{m}%d{d}').format(y='年', m='月', d='日')
            else:
                self.val = dt_datetime.strftime(self.val, str_fmt)
        else:
            raise KeyError("str_typ needs [None,NaN,'str','int','float','datetime','pd_timestamp','tm_structtime']")
        if rtn:
            return self.val

    def dtt_to_prd(self, str_kwd='w', rtn=False, prn=True):
        """
        alter datetime.datetime type to period string in format ['%Yw%w','%Ym%m'].
        >>> dtt = dtz('now')
        >>> dtt.dtt_to_prd('W',rtn=True)
        '2020w17'
        >>> dtt = dtz('now')
        >>> dtt.dtt_to_prd('w',rtn=True)
        '20w17'
        :param str_kwd: in ['W','w','M','m'] for ['%Yw%w','%yw%w','%Ym%m','%ym%m']
        :param rtn: return or not
        :param prn: print or not, default True
        :return: result in format ['%Ym%m','%Yw%w']
        """
        if self.typ not in [dt_datetime]:
            self.typ_to_dtt()
        if self.val is not None:
            slf_dwk = self.val.isocalendar()                        # 得到tuple(year, week, weekday)
            slf_dyr = self.val.timetuple().tm_year                  # 得到year
            slf_dmh = self.val.timetuple().tm_mon                   # 得到month
            int_kyr = slf_dwk[0] if str_kwd.lower() == 'w' else slf_dyr     # 根据str_kwd判断标志年
            int_kwd = slf_dwk[1] if str_kwd.lower() == 'w' else slf_dmh     # 根据str_kwd判断标志字符
            str_kyr = str(int_kyr)[2:] if str_kwd.islower() else str(int_kyr)
            self.val = "%s%s%s" % (str_kyr, str_kwd.lower(), str(int_kwd).zfill(2))
        else:
            if prn:
                print('info: None cannot convert to "%y[wm]%d" format.')
        if rtn:
            return self.val

    def dwk_to_dtt(self, flt_dlt=1, rtn=False):
        """
        alter string in format '19w01' to a certain datetime in this period.
        >>> dtt = dtz('2020w17')
        >>> dtt.dwk_to_dtt(1, rtn=True)
        datetime.datetime(2020, 4, 20, 0, 0)
        >>> dtt = dtz('2020w17')
        >>> dtt.dwk_to_dtt(-1, rtn=True)
        datetime.datetime(2020, 4, 26, 23, 59, 59)
        :param flt_dlt: in range(1,7) for '%Yw%w', range(-7,-1) also equals to range(1,7)
        :param rtn: return the result or not
        :return: if rtn is True, return
        """
        if [True for i in self.fmt if i in ['%Yw%w', '%yw%w']]:
            int_dyr = int(self.val.rsplit('w')[0])
            if int_dyr < 50:
                int_dyr = int(int_dyr + 2000)
            elif int_dyr < 100:
                int_dyr = int(int_dyr + 1900)
            int_dwk = int(self.val.rsplit('w')[1])
            dtt_jan = dt_datetime(int_dyr, 1, 4)
            dtt_dlt = typ_dt_timedelta(dtt_jan.isoweekday() - 1)
            bgn_dyr = dtt_jan - dtt_dlt
            ltr_dlt = typ_dt_timedelta(seconds=60*60*24-1) if flt_dlt < 0 else typ_dt_timedelta(days=0)
            flt_dlt = 8 + flt_dlt if flt_dlt < 0 else flt_dlt
            self.val = bgn_dyr + typ_dt_timedelta(days=flt_dlt-1, weeks=int_dwk-1) + ltr_dlt
        else:
            print('info: {}, {} not in [\'%Yw%w\',\'%yw%w\']'.format(self.val, self.fmt))
        if rtn:
            return self.val

    def dmh_to_dtt(self, flt_dlt=1, rtn=False):
        """
        alter string in format '19m01' to a certain datetime in this period.
        >>> dtt = dtz('19m02')
        >>> dtt.dmh_to_dtt(1,rtn=True)
        datetime.datetime(2019, 2, 1, 0, 0)         # 返回第一天的第一秒
        >>> dtt = dtz('19m02')
        >>> dtt.dmh_to_dtt(-1,rtn=True)
        datetime.datetime(2019, 2, 28, 23, 59, 59)  # 返回最后一天的最后一秒
        :param flt_dlt: in range(1,28/31) for '%Ym%m', range(-31/-28,-1) also equals to range(1,28/31)
        :param rtn: return the result or not
        :return: if rtn is True, return
        """
        if [True for i in self.fmt if i in ['%Ym%m', '%ym%m']]:
            int_dyr = int(self.val.rsplit('m')[0])
            if int_dyr < 50:
                int_dyr = int(int_dyr + 2000)
            elif int_dyr < 100:
                int_dyr = int(int_dyr + 1900)
            int_dmh = int(self.val.rsplit('m')[1])
            int_max = monthrange(int_dyr, int_dmh)[1]
            if int_max - flt_dlt < 0:
                raise KeyError('this month has only %s days' % str(int_max))
            else:
                ltr_dlt = typ_dt_timedelta(seconds=60*60*24-1) if flt_dlt < 0 else typ_dt_timedelta(days=0)
                int_day = flt_dlt if flt_dlt > 0 else int_max + 1 + flt_dlt
                self.val = dt_datetime(int_dyr, int_dmh, int_day) + ltr_dlt
        else:
            print('info: {}, {} not in [\'%Ym%m\',\'%ym%m\']'.format(self.val, self.fmt))
        if rtn:
            return self.val

    def prd_to_dtt(self, flt_dlt=1, rtn=False):
        """
        alter period string in format ['%[Yy]w%w','%[Yy]m%m'] to datetime.datetime type.
        >>> dtt = dtz('19m02')
        >>> dtt.prd_to_dtt(1,rtn=True)
        datetime.datetime(2019, 2, 1, 0, 0)         # 返回第一天的第一秒
        >>> dtt = dtz('19m02')
        >>> dtt.prd_to_dtt(-1,rtn=True)
        datetime.datetime(2019, 2, 28, 23, 59, 59)  # 返回最后一天的最后一秒
        :param flt_dlt: which day in the period to export, in range(1,7) for week and range(1,28-31) for month
        :param rtn: return the result or not
        :return: if rtn is True, return
        """
        if [True for i in self.fmt if i in ['%Yw%w', '%yw%w']]:
            self.dwk_to_dtt(flt_dlt)
        elif [True for i in self.fmt if i in ['%Ym%m', '%ym%m']]:
            self.dmh_to_dtt(flt_dlt)
        else:
            raise AttributeError('%s is not in format [\'%Yw%w\',\'%yw%w\',\'%Ym%m\',\'%ym%m\']')
        if rtn:
            return self.val

    def edg_of_prd(self, lst_flt=None, str_typ='str', str_fmt='%Y-%m-%d', *, rtn=False):
        """
        export the edg of period value in format '[0-9]{2,4}[wm][0-9]{2}'.
        >>> dtz('20w14').edg_of_prd([1, -1], rtn=True)
        ['2020-03-30', '2020-04-05']
        >>> dtz('20w14').edg_of_prd([1, -1], str_fmt='%Y-%m-%d %H:%M:%S', rtn=True)
        ['2020-03-30 00:00:00', '2020-04-05 23:59:59']  # 若lst_flt中的元素为负则时间默认取当天的最后一秒
        :param lst_flt: a list of float in self.val period
        :param str_typ: type of the result
        :param str_fmt: format of the result if type is str
        :param rtn: return the result if True, default False
        :return: return the result if rtn is True
        """
        lst_prd, prd_tmp, lst_flt = [], self.val, [1, -1] if lst_flt is None else lst_flt
        for i in lsz(lst_flt).typ_to_lst(rtn=True):
            self.val = prd_tmp
            self.prd_to_dtt(i)
            lst_prd.append(self.dtt_to_typ(str_typ, str_fmt, rtn=True))
        self.val = lst_prd
        if rtn:
            return self.val

    def shf(self, flt_dlt=0, rtn=False):
        """
        shift days from dtz.val, type of dtz.val will be datetime.datetime.
        >>> dtz('2020-01-01').shf(5, rtn=True)
        datetime.datetime(2020, 1, 6, 0, 0)
        >>> dtz('2020-01-01').shf(-5, rtn=True)
        datetime.datetime(2019, 12, 27, 0, 0)
        :param flt_dlt: how many days to shift, delta in type float
        :param rtn: return or not, default False
        :return: if rtn is True, return the result
        """
        if self.typ not in [dt_datetime]:
            self.typ_to_dtt()
        if self.val is not None:
            self.val += typ_dt_timedelta(days=flt_dlt)
        else:
            print('info: None cannot shift.')
        if rtn:
            return self.val

    def lst_of_day(self, lst_stn, rtn=False):
        """
        generate a list of day.
        >>> dtz().lst_of_day(['2019-01-01','2019-01-04'], rtn=True)
        [datetime.datetime(2019, 1, 1, 0, 0),
         datetime.datetime(2019, 1, 2, 0, 0),
         datetime.datetime(2019, 1, 3, 0, 0)]
        :param lst_stn: a list from start to the end, if lenthg is 1, dtz.val should be the start point
        :param rtn: return or not, default False
        """
        lst_stn = lsz(lst_stn).typ_to_lst(rtn=True)
        if len(lst_stn) == 1:
            lst_stn = [self.val, lst_stn[0]]
        lst_dtt = []
        for i in lst_stn:
            self.val = i
            lst_dtt.append(self.typ_to_dtt(rtn=True))
        self.val = [lst_dtt[0]+typ_dt_timedelta(days=i) for i in range((lst_dtt[1] - lst_dtt[0]).days + 1) if
                    (lst_dtt[0]+typ_dt_timedelta(days=i)).date() != lst_dtt[1].date()]
        if rtn:
            return self.val



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on 2021.05.31
A package slot for ppt generating
requirements = [
  'python-pptx>=0.6.19',
]
@author: zoharslong
"""
from pyzohar.sub_slt_ppt.ppz import ppz

__all__ = ['ppz']
print('info: ppt generating slot inserted.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 22 16:30:38 2019
input/output operation.
@author: zoharslong
"""
from base64 import b64encode, b64decode
from numpy import ndarray as typ_np_ndarray
from pandas.core.series import Series as typ_pd_Series                  # 定义series类型
from pandas.core.frame import DataFrame as typ_pd_DataFrame             # 定义dataframe类型
from pandas.core.indexes.base import Index as typ_pd_Index              # 定义dataframe.columns类型
from pandas.core.indexes.range import RangeIndex as typ_pd_RangeIndex   # 定义dataframe.index类型
from pandas.core.groupby.generic import DataFrameGroupBy as typ_pd_DataFrameGroupBy     # 定义dataframe.groupby类型
from pandas import DataFrame as pd_DataFrame, read_csv, read_excel, concat, ExcelWriter
from time import sleep
from datetime import timedelta as typ_dt_timedelta
from os import listdir, makedirs
from tempfile import gettempdir                                         # 用于搜索fakeuseragent本地temp
from os.path import join as os_join, exists as os_exists
from openpyxl import load_workbook                                      # 保存已有的excel文件中的表
from fake_useragent import UserAgent, VERSION as fku_version            # FakeUserAgentError,
from pymongo import MongoClient
from pymongo.errors import DuplicateKeyError
from pymysql import connect, IntegrityError
from urllib3.util.retry import MaxRetryError
from urllib3.response import ProtocolError
from urllib3.connection import NewConnectionError
from requests.models import ChunkedEncodingError
from requests.adapters import ProxyError
from requests import post, get, TooManyRedirects, ReadTimeout
from re import findall as re_find, sub as re_sub
from random import randint
from json import loads, JSONDecodeError
from pyzohar.sub_slt_bsc.bsz import stz, lsz, dcz, dtz
# from socket import getfqdn, gethostname  # 获得本机IP  # from telnetlib import Telnet  # 代理ip有效性检测的第二种方法


class ioBsc(pd_DataFrame):
    """
    I/O basic
    ioBsc.lcn in {
        'fld','fls',
        'mng','mdb','cln',
        'sql','sdb','tbl',
        'url'/'url_lst'/'url_ctt','url_htp',
        'hdr','pst','prm',
        'prx',‘prx_tms’,
        'ppc':{
            'key': [],
            'ndx': [],
        },
    }
    """
    lst_typ_dts = [
        str,
        stz,
        list,
        lsz,
        dict,
        dcz,
        tuple,
        bytes,
        typ_np_ndarray,
        typ_pd_DataFrame,
        typ_pd_Series,
        typ_pd_Index,
        typ_pd_RangeIndex,
        typ_pd_DataFrameGroupBy,
        type(None)
    ]   # data sets' type
    lst_typ_lcn = [list, lsz, dict, dcz, type(None)]   # io methods' type

    def __init__(self, dts=None, lcn=None, *, spr=False):
        # all the i/o operations have the same attributes for locate target data: location and collection
        super().__init__()                  # 不将dts预传入DataFrame
        self.__dts, self._dts, self.typ = None, None, None   # 接受数据
        self.len, self.clm, self.hdr, self.tal = None, None, None, None
        self.kys, self.vls = None, None
        self.__lcn, self.iot = None, None   # 连接信息
        self._mySql, self._mySdb, self._myTbl = None, None, None
        self._myMng, self._myMdb, self._myCln = None, None, None
        self.__init_rst(dts, lcn, spr=spr)

    def __init_rst(self, dts=None, lcn=None, *, spr=False):
        """
        private reset initiation.
        :param dts: a data set to input or output
        :return: None
        """
        try:
            self.dts = dts.copy() if self.dts is None and dts is not None else []
        except AttributeError:
            self.dts = dts if self.dts is None and dts is not None else []
        self.lcn = lcn.copy() if self.lcn is None and lcn is not None else {}
        if spr:
            self.spr_nit()

    def spr_nit(self, rtn=False):
        """
        super initiation.
        :param rtn: default False
        :return:
        """
        try:
            super(ioBsc, self).__init__(self.__dts)
        except ValueError:
            print('info: %s cannot convert to DataFrame.' % (str(self.__dts)[:8]+'..'))
        if rtn:
            return self

    def __str__(self):
        """
        print(io path).
        :return: None
        """
        dct_prn = {i: self.lcn[i] for i in self.lcn.keys() if i in ['fls', 'cln', 'tbl', 'url']}
        return '<io: %s; ds: %s>' % (str(dct_prn), self.typ)
    __repr__ = __str__  # 调用类名的输出与print(className)相同

    @property
    def dts(self):
        """
        @property get & set lsz.seq.
        :return: lsz.seq
        """
        return self.__dts

    @dts.setter
    def dts(self, dts):
        """
        self.dts = dts
        :param dts: a dataset to import.
        :return: None
        """
        if dts is None or type(dts) in self.lst_typ_dts:
            try:
                self.__dts = dts.copy() if dts is not None else dts
            except AttributeError:
                self.__dts = dts
            self.__attr_rst('dts')
        else:
            raise TypeError('info: dts\'s type %s is not available.' % type(dts))

    def set_dts(self, dts, *, ndx_rst=True, ndx_lvl=None):
        """
        if do not reset index after set data set, use self.set_dts() instead of self.dts
        :param dts: data set to fill self.dts
        :param ndx_rst: if reset data set's index or not, default True
        :param ndx_lvl: DataFrame.reset_index(level=prm), default None
        :return: None
        """
        if dts is None or type(dts) in self.lst_typ_dts:
            try:
                self.__dts = dts.copy() if dts is not None else dts
            except AttributeError:
                self.__dts = dts
            self.__attr_rst('dts', ndx_rst=ndx_rst, ndx_lvl=ndx_lvl)
        else:
            raise TypeError('info: dts\'s type %s is not available.' % type(dts))

    @property
    def lcn(self):
        """
        self.location.
        :return: self.__lcn
        """
        return self.__lcn

    @lcn.setter
    def lcn(self, lcn):
        """
        set self.__lcn in self.lcn.
        :param lcn: a dict of params for self
        :return: None
        """
        if type(lcn) in self.lst_typ_lcn:
            if self.__lcn is None:      # 当self.__lcn为空时, 直接对self__lcn进行赋值
                self.__lcn = lcn
            elif type(lcn) in [dict]:   # 当self.__lcn中已有值时, 使用lcn对其进行更新
                self.__lcn.update(lcn)  # 使用update更新self.__lcn, 要求self.__lcn必为dict类型
            self.__attr_rst('lcn')
        else:
            raise TypeError('info: lcn\'s type %s is not available.' % type(lcn))

    def mng_nit(self):
        """
        if self.io type in mongodb, reset mongo attributes _myMng, _mySdb, _myCln.
        :return: None
        """
        if 'mng' not in self.lcn.keys():
            self.lcn['mng'] = None
        self.lcn['mng'] = "mongodb://localhost:27017" if not self.lcn['mng'] else self.lcn['mng']
        self._myMng = MongoClient(host=self.lcn['mng'])
        self._myMdb = self._myMng[self.lcn['mdb']] if [True if 'mdb' in self.lcn.keys() else False] else None
        self._myCln = self._myMdb[self.lcn['cln']] if [True if 'cln' in self.lcn.keys() else False] else None

    def sql_nit(self):
        """
        SQL initiate. needs self.lcn={'sql'={'hst','prt','usr','psw'},'sdb','tbl'}
        :return: None
        """
        if 'sql' not in self.lcn.keys():
            self.lcn['sql'] = None
        self.lcn['sql'] = {'hst': '172.16.0.13', 'prt': 3306, 'usr': None, 'psw': None} if \
            not self.lcn['sql'] else self.lcn['sql']
        self._mySql = self.lcn['sql'] if [True if 'sql' in self.lcn.keys() else False] else None
        self._mySdb = self.lcn['sdb'] if [True if 'sdb' in self.lcn.keys() else False] else None
        self._myTbl = self.lcn['tbl'] if [True if 'tbl' in self.lcn.keys() else False] else None

    def api_nit(self):
        """
        API initiate. needs self.lcn={'url'/'url_lst'/'url_ctt','pst','hdr','prx','prm'}
        :return:
        """
        # 检查本地fakeUserAgent文件是否存在, 若否则自动创建
        if 'fake_useragent_' + fku_version + '.json' not in listdir(gettempdir()):
            fku = get('https://fake-useragent.herokuapp.com/browsers/' + fku_version, timeout=180)
            with open(os_join(gettempdir(), 'fake_useragent_' + fku_version + '.json'), "w") as wrt:
                wrt.write(fku.text)
        if 'pst' not in self.lcn.keys():
            self.lcn['pst'] = None      # post请求中在请求data中发送的参数数据
        if 'hdr' not in self.lcn.keys():
            self.lcn['hdr'] = {'User-Agent': UserAgent(use_cache_server=False).random}  # 若未指定请求头就现编一个简直可怕
        else:
            self.lcn['hdr'].update({'User-Agent': UserAgent(use_cache_server=False).random})  # 若制定了则自动刷新一次假头
        if 'prx' not in self.lcn.keys():
            self.lcn['prx'] = None      # 是否调用代理
        if 'prm' not in self.lcn.keys():
            self.lcn['prm'] = None      # get请求中后缀于url的参数

    def dts_nit(self, ndx_rst=True, ndx_lvl=None):
        """
        dataset initiate, generate attributes typ, len, kys, vls, clm, hdr, tal and if reset index or not.
        :param ndx_rst: if reset index or not, default True
        :param ndx_lvl: if reset index, set the level of index
        :return: None
        """
        lst_xcp = []
        try:
            self.typ = type(self.__dts)
        except TypeError:
            lst_xcp.append('type')
        try:
            self.len = self.dts.__len__()
        except AttributeError:
            lst_xcp.append('len')
        try:
            self.kys = self.dts.keys()
        except (AttributeError, TypeError):
            lst_xcp.append('keys')
        try:
            self.vls = self.dts.values()
        except (AttributeError, TypeError):
            lst_xcp.append('values')
        if self.typ in [typ_pd_DataFrame]:
            self.clm = self.dts.columns
            self.hdr = self.dts.head()
            self.tal = self.dts.tail()
        self.hdr = self.dts[:5] if self.typ in [list] else self.hdr
        try:
            if ndx_rst:
                self.dts.reset_index(drop=True, inplace=True, level=ndx_lvl)
        except AttributeError:
            lst_xcp.append('resetIndex')
        if not lst_xcp:
            print('info: %s is not available for %s.' % (str(lst_xcp), str(self.__dts)[:8] + '..'))

    def lcn_nit(self, prn=False):
        """
        location initiate, let self.iot in ['lcl','mng','sql','api'] for [local, mongodb, sql, api].
        :return: None
        """
        self.iot = []
        if [True for i in self.lcn.keys() if i in ['fld']] == [True]:
            self.iot.append('lcl')
        if [True for i in self.lcn.keys() if i in ['sdb']] == [True]:
            self.iot.append('sql')
        if [True for i in self.lcn.keys() if i in ['mdb']] == [True]:
            self.iot.append('mng')
        if set([True for i in self.lcn.keys() if re_find('url', i)]) in [{True}]:
            self.iot.append('api')
        if not self.iot and prn:
            print(' info: <.lcn: %s> is not available.' % self.lcn)

    def __attr_rst(self, typ=None, *, ndx_rst=True, ndx_lvl=None):
        """
        reset attributes lsz.typ.
        :param typ: type of attributes resets, in ['dts','lcn'] for data set reset and location reset
        :return: None
        """
        if typ in ['dts', None]:
            self.dts_nit(ndx_rst, ndx_lvl)
        if typ in ['lcn', None]:
            self.lcn_nit()
            if [True for i in self.iot if i in ['mng', 'mnz']]:  # for special cases, reset some attributes
                self.mng_nit()
            if [True for i in self.iot if i in ['sql', 'sqz']]:
                self.sql_nit()
            if [True for i in self.iot if i in ['api', 'apz']]:
                self.api_nit()

    def typ_to_dtf(self, clm=None, *, spr=False, rtn=False):
        """
        self.dts's type from others to dataFrame.
        :param clm: define the columns' name in the final dataFrame
        :param spr: super or not, default False
        :param rtn: return or not, default False
        :return: None if not rtn
        """
        if self.typ in [typ_pd_DataFrame]:
            pass
        elif self.len == 0 or self.dts in [None, [], [{}]]:
            self.dts = pd_DataFrame()
        elif self.typ in [dict, dcz]:
            self.dts = pd_DataFrame([self.dts])
        elif self.typ in [list, lsz, typ_np_ndarray]:
            self.dts = pd_DataFrame(self.dts, columns=clm)
        elif self.typ in [typ_pd_Series]:
            self.dts = pd_DataFrame(self.dts)
        # from sas7bdat import SAS7BDAT as typ_sas7bdat
        # elif self.typ in [typ_sas7bdat]:  # https://pypi.org/project/sas7bdat/
        #     self.dts = self.dts.to_data_frame()
        else:
            raise AttributeError('type of dts is not available')
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def dtf_to_typ(self, typ='list', *, rtn=False, prm='records'):
        """
        alter dataFrame to other type.
        :param typ: alter type dataFrame to this type, default 'list'
        :param rtn: return the result or not, default False
        :param prm: to_dict(orient=''), default 'records' in ['records','dict','list','series','split','index']
        :return: if rtn is True, return the result
        """
        if typ.lower() in ['list', 'lst', 'lsz'] and self.typ in [typ_pd_DataFrame]:
            self.dts = self.dts.to_dict(orient=prm)
        else:
            raise AttributeError('stop: type is not available')
        if rtn:
            return self.dts


class lclMixin(ioBsc):
    """
    local files input and output operations.
    lcn format in {'fld':'','fls':['','',...],'mng':None,'mdb':}.
    >>> lclMixin(lcn={'fld':'dst/samples','fls':['smp01.xlsx']}).lcl_mpt(rtn=True) # 从指定文件夹位置导入文件到内存
       A  B  C
    0  a  1  e
    1  b  2  f
    2  c  3  g
    """

    def __init__(self, dts=None, lcn=None, *, spr=False):
        super(lclMixin, self).__init__(dts, lcn, spr=spr)

    def mpt_csv(self, fld=None, fls=None, sep=None, *, spr=False, rtn=False):
        """
        import csv from folds into RAM.
        :param fld:
        :param fls:
        :param sep:
        :param spr:
        :param rtn:
        :return:
        """
        fld = self.lcn['fld'] if fld is None else fld
        fls = self.lcn['fls'] if fls is None else fls
        fls = fls[0] if type(fls) in [list] else fls
        sep = ',' if sep is None else sep
        self.dts = read_csv(os_join(fld, fls), sep=sep)
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def mpt_xcl(self, fld=None, fls=None, hdr=None, sht=None, *, spr=False, rtn=False):
        """
        import excel data from folds into RAM
        :param fld:
        :param fls:
        :param hdr:
        :param sht:
        :param spr:
        :param rtn:
        :return:
        """
        fld = self.lcn['fld'] if fld is None else fld
        fls = self.lcn['fls'] if fls is None else fls
        fls = fls[0] if type(fls) in [list] else fls
        hdr = 0 if hdr is None else hdr
        sht = 0 if sht is None else sht
        self.dts = read_excel(os_join(fld, fls), header=hdr, sheet_name=sht)
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def mpt_txt(self, fld=None, fls=None, *, spr=False, rtn=False):
        """
        import txt from folds into RAM
        :param fld: target fold
        :param fls: target file
        :param spr:
        :param rtn:
        :return:
        """
        fld = self.lcn['fld'] if fld is None else fld
        fls = self.lcn['fls'] if fls is None else fls
        fls = fls[0] if type(fls) in [list] else fls
        with open(os_join(fld, fls), mode='r', encoding='utf-8') as act:
            self.dts = act.readlines()  # 读出的为列表
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def mpt_img(self, fld=None, fls=None, *, spr=False, rtn=False):
        """
        import image from disc
        :param fld: target fold
        :param fls: target file
        :param spr:
        :param rtn:
        :return: if rtn, return a string in type base64
        """
        fld = self.lcn['fld'] if fld is None else fld
        fls = self.lcn['fls'] if fls is None else fls
        fls = fls[0] if type(fls) in [list] else fls
        with open(os_join(fld, fls), mode='rb') as act:
            self.dts = b64encode(act.read())
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def lcl_mpt(self, *, sep=None, hdr=None, sht=None, spr=False, rtn=False):
        """
        local files importation.
        :param sep:
        :param hdr:
        :param sht:
        :param spr:
        :param rtn:
        :return:
        """
        if type(self.lcn['fls']) is str:    # 对可能存在的'fls'对多个文件的情况进行统一
            self.lcn = {'fls': [self.lcn['fls']]}
        if type(self.dts) in [typ_pd_DataFrame] and self.len == 0:
            dtf_mrg = pd_DataFrame()
        elif self.dts is None or self.dts == []:
            dtf_mrg = pd_DataFrame()
        else:
            dtf_mrg = self.dts.copy()
        for i_fls in self.lcn['fls']:
            if i_fls.rsplit('.')[1] in ['csv']:
                self.mpt_csv(fls=i_fls, sep=sep)
            elif i_fls.rsplit('.')[1] in ['xls', 'xlsx']:
                self.mpt_xcl(fls=i_fls, hdr=hdr, sht=sht)
            elif i_fls.rsplit('.')[1] in ['txt']:
                self.mpt_txt(fls=i_fls)
                self.typ_to_dtf()
            elif i_fls.rsplit('.')[1] in ['jpeg', 'jpg', 'png']:
                self.mpt_img(fls=i_fls)
                self.dts = pd_DataFrame([self.dts], columns=[i_fls.rsplit('.')[1]])
            dtf_mrg = concat([dtf_mrg, self.dts], ignore_index=True)    # 忽视index的多文件纵向拼接, prm: sort=False
        self.dts = dtf_mrg
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def xpt_txt(self, fld=None, fls=None, *, typ='w', sep=2, cvr=True):
        """
        import dataset to file.txt or file.js(https://blog.csdn.net/matrix_google/article/details/76861485).
        :param fld: fold location
        :param fls: files name
        :param typ: type of open, usually in ['a','w'], a for continue, w for cover
        :param sep: cut how many units in the head and tail of each row, default 2 to be compatible with echarts
        :param cvr: check if the pth_txt is already exists or not, False means if it exists, then do nothing
        :return: None
        """
        fld = self.lcn['fld'] if fld is None else fld
        fls = self.lcn['fls'] if fls is None else fls
        fls = fls[0] if type(fls) in [list] else fls
        if os_exists(os_join(self.lcn['fld'], fls)) and cvr is False:
            print("stop: the txt %s already exists." % str(self.lcn.values()))
        elif type(self.dts) not in [str, list, dict, typ_pd_DataFrame]:
            print('stop: type of dts_npt needs [str, list, dict, pd.DataFrame].')
        else:
            self_dts = [self.dts] if self.typ in [str, dict] else self.dts
            if self.typ in [str, dict, list]:
                lst_dts_mpt = [str(i_dcm) for i_dcm in self_dts]
            else:  # alter the type of a line in dataframe from slice to str
                lst_dts_mpt = [str(i_dcm)[sep: -sep] for i_dcm in self_dts.to_dict(orient='split')['data']]
            if not os_exists(self.lcn['fld']):
                makedirs(self.lcn['fld'])
            prc_txt_writer = open(os_join(fld, fls), typ, encoding='utf-8')
            for i in range(len(self_dts)):
                if typ in ['a', 'A']:
                    prc_txt_writer.write('\n')
                prc_txt_writer.writelines(lst_dts_mpt[i])
            prc_txt_writer.close()
            print("info: success input dts to txt %s." % str(self.lcn.values()))

    def lcl_xpt(self, *, typ='w', sep=2, cvr=True, ndx=False, prm='sheet1', fld=None, fls=None):
        """
        local exporting, type in lcn['fls']
        :param typ: exporting type in ['w','a'] if lcn['fls'] in ['.js','.txt'], default 'w' means cover the old file
        :param sep: cut how many units in the head and tail of each row, default 2 to be compatible with echarts
        :param cvr: check if the pth_txt is already exists or not, False means if it exists, then do nothing
        :param ndx: remain index in the first line if lcn['fls'] in ['.xlsx'], default False
        :param prm: if typ='a' and fls in type '.xlsx', continue to write the excel, use prm to define the sheet name
        :param fld: default None, use self.lcn['fld']
        :param fls: default None, use self.lcn['fls']
        :return: None
        """
        if fld is None:
            fld = self.lcn['fld']
        if fls is None:
            fls = self.lcn['fls'][0] if type(self.lcn['fls']) in [list] else self.lcn['fls']
        if fls.rsplit('.')[1] in ['xlsx'] and typ in ['w'] and cvr:
            self.typ_to_dtf()
            self.dts.to_excel(os_join(fld, fls), sheet_name=prm, index=ndx)
        elif fls.rsplit('.')[1] in ['xlsx'] and (typ in ['a'] or not cvr):
            self.typ_to_dtf()
            writer = ExcelWriter(os_join(fld, fls), engine='openpyxl')
            try:
                book = load_workbook(os_join(fld, fls))     # 尝试保存已有的文件内容
                writer.book = book
            except FileNotFoundError:
                pass
            self.dts.to_excel(writer, sheet_name=prm, index=ndx)
            writer.save()
            writer.close()
        elif fls.rsplit('.')[1] in ['csv']:
            self.typ_to_dtf()
            self.dts.to_csv(os_join(fld, fls), encoding='UTF-8_sig')    # 不明原因的解码方式
        elif fls.rsplit('.')[1] in ['js', 'txt']:
            self.xpt_txt(typ=typ, sep=sep, cvr=cvr)
        elif fls.rsplit('.')[1] in ['png', 'jpeg', 'jpg']:              # 图片文件导出到硬盘
            if type(self.dts) in [bytes, str]:
                self_dts = b64decode(self.dts)
            elif type(self.dts) in [typ_pd_DataFrame]:
                self_dts = b64decode(self.dts.iloc[0, 0])
            else:
                self_dts = b64decode(self.dts[0])
            with open(os_join(fld, fls), 'wb') as act:
                act.write(self_dts)
        else:
            print('stop: known exporting type.')


class mngMixin(ioBsc):
    """
    local files input and output operations.
    >>> from pandas import DataFrame
    >>> tst = ioz(dts=DataFrame([{'A':'a','B':1.0},{'A':'b','B':2.0}]),lcn={'mdb':'db_tst','cln':'cl_tst'})
    >>> tst.mng_xpt()
    ***** db_tst . cl_tst *****
    info: 2 inserted, 0 updated, 0 dropped.
    >>> tst.mng_mpt(rtn=True)
    [{'_id': ObjectId('5e48f0ae8daa9bfcc3c0987f'), 'A': 'a', 'B': 1.0},
     {'_id': ObjectId('5e48f0ae8daa9bfcc3c09880'), 'A': 'b', 'B': 2.0}]
    """
    def __init__(self, dts=None, lcn=None, *, spr=False):
        super(mngMixin, self).__init__(dts, lcn, spr=spr)
        self._mpt_id = None

    def mng_nfo(self, typ=None, *, prm=None, dct_qry=None, lst_clm=None):
        """
        export mongodb connection information.
        >>> dfzMng = mngMixin(lcn={'mdb':'dbs_tst','cln':'cln_tst'})
        >>> dfzMng.mng_nfo('max', prm='x_tst') # 返回prm列的最大值
        '20w13'
        :param typ: which information is needed, in ['dbs','cln','clm','dcm','vls','max','min']
        :param prm: if typ is 'dcm', insert a certain column name for all the unique values in this column
        :param dct_qry: only referred when typ in ['vls','max','min'] for documents' choosing
        :param lst_clm: only referred when typ in ['vls','max','min'] for columns' choosing
        :return: a list of target values
        """
        if typ in ['dbs', 'database', 'mdb', 'mongodatabase']:
            return self._myMng.list_database_names()
        elif typ in ['cln', 'collection']:
            return self._myMdb.list_collection_names()
        elif typ in ['clm', 'column', 'columns']:
            return list(self._myCln.find_one().keys())
        elif typ in ['dcm', 'document', 'documents']:
            return self._myCln.distinct(prm)
        elif typ in ['vls', 'values', 'max', 'min']:
            flt_typ = -1 if typ in ['max'] else 1
            dct_qry = {} if not dct_qry else dct_qry
            dct_clm = dcz().typ_to_dct(lst_clm, 1, rtn=True) if lst_clm else None
            try:                # 当对目标collection的find无异常时
                mtx_stt = self._myCln.find(dct_qry, dct_clm).sort([(prm, flt_typ)]).limit(100)
                lst_xpt = []
                for dct_xpt in mtx_stt:
                    if dct_xpt not in lst_xpt:
                        lst_xpt.extend([dct_xpt])
                if lst_xpt:
                    return lst_xpt[0][prm]
                else:           # 当目标collection的find未找到document时返回空
                    return None
            except KeyError:    # 当目标collection导出失败时返回空
                return None
        else:
            return [self._myMdb.name, self._myCln.name, list(self._myCln.find_one().keys())]

    def crt_ndx(self, lst_ndx, unq=False, drp=True, srt=1):
        """
        当选定unique indexs时，索引在本集合中唯一，重复插入报错KeyError
        :param lst_ndx: 索引列表，也可以为想要添加为复合索引的列名的list, 最终自动转化为[(,1),(,1),...]形式
        :param unq: 单个索引或复合索引在本集合中是否唯一，若是则重复插入的同索引将报错KeyError，默认False为可重复
        :param drp: 在生成新索引前是否删除全部现有的索引，默认True为全部删除
        :param srt: 建立索引时各列名为正序（1）或倒序（-1），默认为1即全部正序
        @param lst_ndx:
        @param unq:
        @param drp:
        @param srt:
        """
        lst_ndx = lsz(lst_ndx).lst_to_typ('listtuple', srt, rtn=True)   # return [('A',1),('B',1)]
        if drp:
            self._myCln.drop_indexes()                                  # if True, delete old indexes
        self._myCln.create_index(lst_ndx, unique=unq)

    def dlt_cln(self, ask=True):
        """pymongo.drop
        print 'Y' to make sure that the target collection should be dropped.
        :return: None
        @param ask:
        """
        if ask:
            if input("delete(y/n): %s\n" % str(self._myCln)).lower() in ["y", "yes", "t", "true"]:
                print("info: deleted: %s\n" % str(self._myCln))
                self._myCln.drop()
            else:
                raise KeyError('stop: nothing happened for unknown input(needs yes/true)')
        else:
            self._myCln.drop()

    def dlt_cln_dcm(self, dct_qry=None, *, rtn=False):
        """

        @param dct_qry:
        @param rtn:
        @return:
        """
        if dct_qry is None:
            if input("delete all(y/n): %s\n" % str(self._myCln)).lower() not in ["y", "yes", "t", "true"]:
                raise KeyError('nothing will happen for null dct_qry and wrong answer')
        prc_delete_many = self._myCln.delete_many(dct_qry)
        if rtn:
            return prc_delete_many.deleted_count

    def ltr_cln_dcm(self, dct_qry=None, dct_ltr=None):
        """

        @param dct_qry:
        @param dct_ltr: needs '$set', '$inc'
        @return:
        """
        prc_update_many = self._myCln.update_many(dct_qry, dct_ltr, upsert=True)
        if prc_update_many.raw_result['updatedExisting']:
            return prc_update_many.raw_result

    def mng_mpt(self, dct_qry=None, lst_clm=None, *, prm='', spr=False, rtn=False):
        """
        import data from mongodb to RAM.
        :param dct_qry: a dict of query
        :param lst_clm: default not import _id
        :param prm: match 'onepiprcrdtf' for find_one, aggregate, return recuration, return dataframe
        :param spr: up to self, default False
        :param rtn: return the result or not, default False
        :return: None
        """
        dct_qry = {} if not dct_qry else dct_qry
        dct_clm = {'_id': 0}
        if lst_clm is not None:
            dct_clm.update(dcz().typ_to_dct(lst_clm, 1, rtn=True))
        if re_find('one', prm):
            prc_find = [self._myCln.find_one(dct_qry, dct_clm)]
        elif re_find('pip', prm):
            prc_find = self._myCln.aggregate(dct_qry)
        else:
            prc_find = self._myCln.find(dct_qry, dct_clm)
        if re_find('rcr', prm):     # prm in ['rcr']直接返回迭代器
            return prc_find
        else:                       # 否则将全部documents解出存放再self.dts中
            self.dts = [dct_xpt for dct_xpt in prc_find] if prc_find not in [None, [None]] else []
            if prm in ['dtf', 'dataframe']:
                self.typ_to_dtf()
            if spr:
                self.spr_nit()
            if rtn:
                return self.dts

    def xpt_cln_dcm(self, dts_mpt, lst_ndx=None, cvr=True, rtn=False):
        """
        向集合中更新一行pd.Series，若lst_clm_ndx重复则可以选择是否覆盖
        :param dts_mpt: 一行数据，进入集合后视为一个文档，可选类型为[pd.Series, dict]
        :param lst_ndx: 在此集合中选定为索引且unique=True的列名
        :param cvr: 当索引重复时是否用dts_mpt覆盖集合内的行，默认True
        :param rtn: 返回输入的情况
        :return: None
        """
        self._mpt_id = [] if not self._mpt_id else self._mpt_id     # 为后续记录输入文档的id做准备
        if type(dts_mpt) == typ_pd_Series:
            lst_ndx = dts_mpt.index.to_list() if lst_ndx is None else lst_ndx
            dct_mpt = dts_mpt.to_dict()
        elif type(dts_mpt) == dict:
            lst_ndx = list(dts_mpt.keys()) if lst_ndx is None else lst_ndx
            dct_mpt = dts_mpt.copy()
        else:
            raise KeyError("stop: type of dts_mpt needs in [pd.Series, dict]")
        n, c, d = 0, 0, 0
        dct_ndx = {i: dct_mpt[i] for i in lst_ndx}  # 构造与索引同格式的用于核实是否已存在于库中的列字典
        try:
            prc_insert = self._myCln.insert_one(dct_mpt)
            self._mpt_id.append(prc_insert.inserted_id)
            n += 1
        except DuplicateKeyError:
            dct_mpt.pop('_id')
            if cvr and len(dct_mpt) > 0:
                self._myCln.update_one(dct_ndx, {'$set': dct_mpt})
                c += 1
            else:
                d += 1
        if rtn:
            return n, c, d

    def xpt_cln_dtf(self, lst_ndx=None, cvr=True):
        """
        逐行更新，保持集合中lst_ndx唯一
        :param lst_ndx: 在此集合中选定为索引且unique=True的列名
        :param cvr: 当索引重复时是否用dts_mpt覆盖集合内的行，默认True
        :return: None
        """
        self.typ_to_dtf()
        nt, ct, dt = 0, 0, 0
        for i in range(self.len):
            n, c, d = self.xpt_cln_dcm(self.dts.loc[i, :], lst_ndx, cvr, True)
            nt += n
            ct += c
            dt += d
        print('info: %i inserted, %i updated, %i dropped.' % (nt, ct, dt))

    def mng_xpt(self, lst_ndx=None, cvr=True):
        """import from dataframe to collections
        根据库是否存在, 进行有选择性的导入策略,全部导入或更新导入
        :param lst_ndx: 为新建的集合创建唯一索引组
        :param cvr: 覆盖或是舍弃, default cover, True
        :return: None
        """
        print('*****', self._myMdb.name, '.', self._myCln.name, '*****')
        # estimated_document_count/count_documents instead count
        if self._myCln.estimated_document_count() == 0:
            self.xpt_cln_dtf()
            if lst_ndx is not None:
                self.crt_ndx(lst_ndx, True)
        else:
            self.xpt_cln_dtf(lst_ndx, cvr)


class sqlMixin(ioBsc):
    """
    mysql tables input and output operations.
    >>> tst = ioz(lcn={'sql':{'hst':"cdb-********.gz.tencentcdb.com", 'prt':1***9,'usr':'**','psw':'**********'},
    >>>                'sdb':'db_spd_dly',
    >>>                'tbl':'tb_spd_est_shd_ljw_190409',})
    >>> tst.sql_run("SELECT table_name FROM information_schema.tables WHERE table_schema=%s", tst.lcn['sdb'])  # 查看表名
    """
    def __init__(self, dts=None, lcn=None, *, spr=False):
        super(sqlMixin, self).__init__(dts, lcn, spr=spr)

    def sql_run(self, str_sql=None, lst_kys=None, *, spr=False, rtn=False):
        """
        tencent cloud mysql operation
        # create
        str_clm = 'estatename CHAR(36) PRIMARY KEY, ndx TINYINT NOT NULL'
        sql_crt = "CREATE TABLE %s(%s)ENGINE=innodb DEFAULT CHARSET=utf8;"%('tb_spd_bdu_adv_kwd',str_clm)
        # alter - add multi primary key
        sql_ltr = 'ALTER TABLE %s ADD PRIMARY KEY(do_date, estatename, _cty)'%('tb_spd_bdu_adv')
        # insert
        sql_nst = 'insert into %s(%s) values(%s)'%('tb_spd_bdu_adv_kwd','estatename, ndx','"云", "5"')
        # update
        sql_pdt = 'update %s set %s = %s where %s = %s' %(
            'tb_spd_bdu_adv_kwd','estatename','"桑泰龙樾"','estatename','"桑泰龙樾"')
        # delete
        sql_dlt = 'delete from %s where %s=%s'%('tb_spd_bdu_adv_kwd','estatename','"云"')
        # select
        sql_slt = 'select * from %s'%('tb_spd_bdu_adv_kwd')
        # drop
        sql_drp = 'DROP TABLE %s'%('tb_spd_adv_bdu_kwd')
        # show tables' name in database
        >>> chk = "SELECT table_name FROM information_schema.tables WHERE table_schema='%s'" % self.lcn['sdb']
        >>> sql = sqlMixin()    # 以下演示execute中sql语句的两种构造方法, 两种方法效果相同
        >>> sql.sql_run('SELECT * FROM %s WHERE %s' % (sql.lcn['tbl'], 'id>="100"'))      # 在crs.execute中只使用sql参数
        >>> sql.sql_run("SELECT * FROM %s WHERE id>= " % sql.lcn['tbl'] + '%s', '100')    # 使用了只能定义value的args参数
        @param str_sql: sql sentences
        @param lst_kys: in crs.execute(str_sql, lst_kys) for %s without '"', only can use in WHERE's value part
        @param spr: let ioz = ioz.dts or not, default False
        @param rtn: return the result or not, default False
        @return: the target dataset in type DataFrame or None
        """
        cnn = connect(host=self._mySql['hst'], port=self._mySql['prt'],
                      user=self._mySql['usr'], password=self._mySql['psw'],
                      database=self._mySdb, charset="utf8")
        crs = cnn.cursor()
        try:
            crs.execute(str_sql, lst_kys)
            if str_sql[:6].lower() in ['select']:
                dts = crs.fetchall()    # values in type tuple
                clm = crs.description   # columns information
                self.dts = pd_DataFrame(list(dts), columns=[i_clm[0] for i_clm in list(clm)]) if dts else self.dts
            elif str_sql[:6].lower() in ['insert', 'update', 'delete', 'replac']:
                cnn.commit()
            elif str_sql[:6].lower() in ['create']:
                pass
            else:
                raise KeyError('stop: the sentence of sql needs [select, insert, update, delete]')
            if spr:
                self.spr_nit()
            if rtn:
                return self.dts
        except IntegrityError:
            return 'stop: pymysql.IntegrityError'
        except ():
            return 'stop: unknown error.'
        finally:
            crs.close()
            cnn.close()

    def sql_xpt(self):
        """
        export self.dts into mysql, insert and update when unique keys error
        """
        self.typ_to_dtf()
        nt, ct, dt = 0, 0, 0
        str_clm = re_sub("'", '', str(self.clm.tolist()))[1:-1]
        for i_ndx in range(self.len):
            str_ctt = '"'
            for i in range(len(self.clm)):
                str_ctt += self.dts.iloc[i_ndx, :][i]
                str_ctt += '","'
            str_ctt = str_ctt[:-2]
            nt += 1
            rsl = self.sql_run('insert into %s(%s) values(%s)' % (self.lcn['tbl'], str_clm, str_ctt))
            if rsl and re_find('IntegrityError', rsl):
                self.sql_run('replace into %s(%s) values(%s)' % (self.lcn['tbl'], str_clm, str_ctt))
                ct += 1
                nt -= 1
        print('info: %s, %i inserted, %i updated.' % (self.lcn['tbl'], nt, ct))

    def sql_mpt(self, str_qry=None, lst_clm='*', *, spr=False, rtn=False):
        """
        Import data from mysql
        :param str_qry: a string after 'WHERE', in format 'columnName>="value"'
        :param lst_clm: choose the columns' name to import
        :param spr: let ioz = ioz.dts or not, default False
        :param rtn: return the result or not, default False
        """
        str_sql = 'SELECT %s FROM %s' % (re_sub('[\[\]\'\"]', '', str(lst_clm)), self.lcn['tbl'])
        if str_qry is not None:
            str_sql = str_sql + ' WHERE ' + str_qry
        self.sql_run(str_sql)
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts


class apiMixin(ioBsc):
    """
    local files input and output operations.
    >>> self = ioz(lcn={
    >>>     'url':"http://haixfsz.centaline.com.cn/service/postda",
    >>>     'url_htp': 'post',  # 用于制定调用接口的http方法, 默认post, 也可以于此处指定或在self.api_run(prm='post')指定
    >>>     'prx': 'auto',      # 是否开启代理
    >>>     'prx_tms': '%Y-%m-%d %H:%M:%S',         # 更新代理的最后时间
    >>>     'prx_tkn': {'neek':'','appkey':''},     # 极光代理的token
    >>>     'pst':{
    >>>         'uid':'haixf**',
    >>>         'pwd':'**.***$',
    >>>         'key':'requestCCESdeallog',
    >>>         'ContractDate_from':'2020-01-01',
    >>>         'ContractDate_to':'2020-01-02'
    >>>     }, # 用于post方法传参
    >>>     'prm':{},  # 用于get方法传参
    >>>     'hdr':{},
    >>>     'ppc': {
    >>>         'key': [],      # 从返回数据中提取目标数据的key
    >>>         'xpr': 0,       # 设定某个刷新本数据的超时秒数
    >>>     }
    >>> })
    >>> self.api_run()  # 从api导入数据
    >>> self.mng_xpt()  # 将api数据导出到mongodb
    """
    def __init__(self, dts=None, lcn=None, *, spr=False):
        super(apiMixin, self).__init__(dts, lcn, spr=spr)

    def _pi_prx_jgw(self, prm='http', rty=3, dct_wht=None):
        """
        get proxy from jiguang api. 从极光代理获取一个代理ip地址, 同时使用self.lcn['prx_tms']记录这个代理的获取时间
        from http://h.jiguangdaili.com/api/new_api.html; userid 158xxxx9977;
        :param prm: in ['http', 'https']
        :param rty: retry times, default 3
        :param dct_wht: 极光代理的卡密白名单, http://h.jiguangdaili.com/api/new_api.html; userid 158xxxx9977;
        :return: None
        """
        dct_jgw = {
            'url': 'http://d.jghttp.golangapi.com/getip',
            'prm': {
                'num': '1',         # 一次请求的数量
                'type': '2',        # json type
                'pro': '0',         # 440000 - 广东
                'city': '0',        # 440100 - 广州; 441900 - 东莞
                'yys': '0',
                'port': '1',        # 1 for http, 11 for https
                'time': '1',        # [1,2,3] for different available time period
                'ts': '0',
                'ys': '0',
                'cs': '0',
                'lb': '1',
                'sb': '0',
                'pb': '4',
                'mr': '1',
                'regions': '440000',  # 广东省份混拨
                # 'pack': '20855',    # 套餐ID, 无套餐直接使用余额则为空
            },
            'ppc': {'key': ['data']},
        }
        dct_jgw['prm']['port'] = '1' if prm == 'http' else '11'
        dct_wht = self.lcn['prx_tkn'] if 'prx_tkn' in self.lcn.keys() else dct_wht  # 从self.lcn['prx_tkn']获得代理token
        htp = 'http' if prm == 'http' else 'https'
        prc, bch, ipp_prx = True, 0, None
        while prc and bch <= rty:
            if bch == rty:                              # 当运行到第四轮时, 视为运行失败, 直接报错跳出
                raise AttributeError('stop: something wrong with jiguang API, max retry.')
            mdl_prx = get(dct_jgw['url'], params=dct_jgw['prm'], timeout=180)
            mdl_prx.encoding = "utf-8"
            dct_prx = loads(mdl_prx.text)
            try:                                        # 尝试拼接标准格式的proxy
                ipp_prx = htp + '://' + str(dct_prx['data'][0]['ip']) + ':' + str(dct_prx['data'][0]['port'])
                print('info: proxy shifted to %s.' % ipp_prx)
                prc = False
            except (IndexError, KeyError):              # 当API未能返回有效格式的proxy时，尝试将当前机器IP添加白名单来解决这个问题
                if dct_prx['code'] in [113, '113']:     # {"code":113,"data":[],"msg":"请添加白名单xxxx","success":false}
                    url_wht = 'http://webapi.jghttp.golangapi.com/index/index/save_white'
                    dct_wht.update({'white': re_find('请添加白名单(.*$)', dct_prx['msg'])[0]})
                    rps_wht = get(url_wht, params=dct_wht)
                    print(rps_wht.text)
                    sleep(3)                            # 添加白名单的操作后需要等待2秒
                elif dct_prx['code'] in [111, '111']:   # {"code":111,"data":[],"msg":"请2秒后再试","success":false}
                    sleep(int(re_find('请(\d+)秒后再试', dct_prx['msg'])[0])+1)
                else:                                   # 当API异常并不是由于白名单问题造成时报错
                    raise KeyError('stop: something wrong with the result <%s>.' % str(mdl_prx.text))
            finally:
                bch += 1
        if ipp_prx:
            self.lcn['prx'] = {htp: ipp_prx}
            self.lcn['prx_tms'] = dtz('now').typ_to_dtt(rtn=True)                               # 定时更换proxy的时间戳
            self.lcn['hdr'].update({'User-Agent': UserAgent(use_cache_server=False).random})    # 在更新proxy的同时更换头

    def _pi_ipl(self, url="http://icanhazip.com/"):
        """
        返回本机当次的ip地址（若self.lcn.prx存在则返回该代理的ip地址）. get local ip from API.
        from https://blog.csdn.net/weixin_44285988/article/details/102837864
        from https://www.cnblogs.com/hankleo/p/11771682.html
        # >>> Telnet('116.22.50.144', '4526', timeout=2)
        <telnetlib.Telnet at 0x1e8dcd8d548>     # 返回一个实例化后的类，说明该代理ip有效，否则返回timeOutError
        :param url: 用于返回本次使用的本机或代理ip的API地址
        :return: ip location in str
        """
        if self.lcn['prx'] in ['auto']:
            raise AttributeError('stop: self.lcn.prx is auto, proxy does not available yet. run api_prx before this.')
        try:
            mdl_rqt = get(url, proxies=self.lcn['prx'], timeout=180)
            if mdl_rqt.status_code == 200:
                return mdl_rqt.text.replace('\n', '')
            elif mdl_rqt.status_code in [402, 403]:
                return re_find("IP\<\/span>: ([0-9.]*?)<", mdl_rqt.text)[0]
            elif mdl_rqt.status_code == 502:
                return 'stop: 502 - connection timed out.'
            else:
                return mdl_rqt
        except (MaxRetryError, NewConnectionError, ConnectionResetError, ProxyError, ReadTimeout):
            return 'stop: MaxRetryError/NewConnectionError.'

    def _pi_prx_chk(self):
        """
        API: if proxies is alright, return True, else return False.
        # >>> Telnet('116.22.50.144', '4526', timeout=2)
        # <telnetlib.Telnet at 0x1e8dcd8d548>     # 返回一个实例化后的类，说明该代理ip有效，否则返回timeOutError
        # >>> # 如上方法的脚本示例
        # >>> try:
        # >>>     if Telnet(re_find('http://(.*?):.*$', self.lcn['prx'][list(self.lcn['prx'].keys())[0]])[0],
        # >>>               re_find('http://.*:(.*$)', self.lcn['prx'][list(self.lcn['prx'].keys())[0]])[0],
        # >>>               timeout=30):  # 使用telnet验证代理效果
        # >>>         return True
        # >>> except TimeoutError:
        # >>>     return False
        :return: True or False for a good proxy
        """
        if self.lcn['prx'] in [None, 'auto']:
            return False
        rsp = self._pi_ipl()        # 可能返回字符串, <respons401> 等
        if type(rsp) not in [str]:
            return False            # 非字符串默认无效
        elif re_find(rsp, self.lcn['prx'][list(self.lcn['prx'].keys())[0]]):
            return True             # 当可以在ip地址测试中发现本次挂载的代理ip时则证明代理成功
        else:
            return False            # 其他情况默认无效

    def api_prx(self, prm='http', frc=False, rty=5, dct_jgh=None):
        """
        当self.lcn.prx代理不为空, 即采用代理时, 调用代理是否有效的检查，若无效则更新代理, 更新五次失败则报错
        :param frc: forced refresh, default False means not refresh if ip test is passed
        :param prm: in ['http','https'], default 'http'
        :param rty: retry times, default 2
        :param dct_jgh: 极光代理的卡密, http://h.jiguangdaili.com/api/new_api.html; userid 158xxxx9977;
        :return:
        """
        if self.lcn['prx']:
            bch, dtz_now = 0, dtz('now').typ_to_dtt(rtn=True)
            if frc:
                self._pi_prx_jgw(prm=prm, dct_wht=dct_jgh)       # 强制更新一次proxy
            elif 'prx_tms' in self.lcn.keys() and (dtz_now - self.lcn['prx_tms']).seconds > randint(0, 30) + 60:
                self._pi_prx_jgw(prm=prm, dct_wht=dct_jgh)       # 原代理超时后强制更新一次proxy
            while not self._pi_prx_chk() and bch <= rty:
                if bch == rty:
                    raise AttributeError('stop: max retry, cannot pass proxy test for %s times.' % str(rty))
                self._pi_prx_jgw(prm=prm, dct_wht=dct_jgh)
                bch += 1

    def api_run(self, *, spr=False, rtn=False, prm='post', frc=False, rty=3, dct_jgh=None):
        """
        兼容POST/GET两种方法的调用, self.lcn.pst在post中输入data, self.lcn.prm在get中输入params.
        此处的循环检查只针对网页连接结果反馈不为200的情况, 而不是检查返回的内容是否满足需求.
        from https://www.cnblogs.com/roadwide/p/10804888.html
        :param spr:
        :param rtn:
        :param prm: in ['post', 'get'], the methods of requests, self.lcn['url_htp']
        :param frc: force change proxy or not
        :param rty: retry, if the result is not available, define retry times, default 3.
        :param dct_jgh: 极光代理的卡密, http://h.jiguangdaili.com/api/new_api.html; userid 158xxxx9977;
        :return:
        """
        if 'url' not in self.lcn.keys():
            raise KeyError('stop: url not exist.')
        else:
            self.api_prx(frc=frc, dct_jgh=dct_jgh)  # 初始化更新proxy, 仅用于处理prx='auto'的情况
            prc, bch, mdl_rqt, rty = True, 0, None, 1 if self.lcn['prx'] in [None] else rty
            prm = self.lcn['url_htp'] if 'url_htp' in self.lcn.keys() else prm  # self.lcn优先
            while prc and bch <= rty:
                # 满足本次请求的返回statusCode为200即请求成功, 或循环rty次失败
                try:
                    mdl_rqt = post(
                        self.lcn['url'], params=self.lcn['prm'], data=self.lcn['pst'],
                        headers=self.lcn['hdr'], proxies=self.lcn['prx'], timeout=300
                    ) if prm.lower() in ['post', 'pst'] else get(
                        self.lcn['url'], params=self.lcn['prm'], data=self.lcn['pst'],
                        headers=self.lcn['hdr'], proxies=self.lcn['prx'], timeout=300
                    )                   # 针对POST/GET请求, 分别注意self.lcn.pst/self.lcn/prm参数
                    mdl_rqt.encoding = "utf-8"
                    if mdl_rqt.status_code in [200]:
                        prc = False             # 当不使用代理或网页返回了信息时, 不进行循环
                    else:
                        self.api_prx(frc=frc, dct_jgh=dct_jgh)   # 其他情况酌情更新代理
                except (
                        ProtocolError, ChunkedEncodingError, TooManyRedirects,
                        MaxRetryError, NewConnectionError, ConnectionError, TimeoutError
                ):   # proxy ip timeout, change proxy
                    print('info: ConectionError, retrying.')
                    sleep(25+randint(0, 5))
                    self.api_prx(frc=frc, dct_jgh=dct_jgh)
                finally:
                    bch += 1
                    if bch == rty:              # 当不使用代理时，默认的rty=1
                        self.api_prx(frc=True, dct_jgh=dct_jgh)  # 当本次POST/GET请求未能得到200且为最后一次运行时, 代理强制更新
                        print('info: max batches, proxy forces switch.')
            # 请求阶段结束, 进入API结果的装载阶段
            try:                                # 优先尝试js解码
                self.dts = loads(mdl_rqt.text)
            except JSONDecodeError:             # 若js解码失败则直接装载self.dts
                self.dts = mdl_rqt.text
            if spr:
                self.spr_nit()
            if rtn:
                return self.dts

    def get_vls(self, lst_kys=None, *, spr=False, rtn=False):
        """
        get values from dict by keys in api feedback.
        :param lst_kys:
        :param spr:
        :param rtn:
        :return:
        """
        if lst_kys:
            lst_kys, i = lsz(lst_kys).typ_to_lst(rtn=True), 'initiation'
            try:
                for i in lst_kys:
                    self.dts = self.dts[i]
                if spr:
                    self.spr_nit()
                if rtn:
                    return self.dts
            except (KeyError, AttributeError):
                print(str(self.dts)[:32]+'..')
                raise KeyError('%s do not exist, something wrong with this request.' % i)

    def api_mpt(self, lst_kys=None, *, spr=False, rtn=False, prm='post', frc=False, rty=3, dct_jgh=None):
        """
        import data from API in json decoding.
        :param lst_kys:
        :param spr:
        :param rtn:
        :param prm:
        :param frc: shift proxy ip forced
        :param rty: retry times, default 3
        :param dct_jgh: 极光代理的卡密, http://h.jiguangdaili.com/api/new_api.html; userid 158xxxx9977;
        :return:
        """
        self.api_run(prm=prm, frc=frc, rty=rty, dct_jgh=dct_jgh)
        self.get_vls(lst_kys)
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def api_tkn(self, dly=3600, frc=False, rty=2):
        """
        Get API token from wechat, baiduMap, etc. Needs lcn.fls in type txt.
        >>> # for a API token class, we need keys below
        >>> lcn_bdu_ocr_tkn = {
        >>>     'fld': './dst/doc_tkn', 'fls': 'tkn_bdu_ocr.txt',
        >>>     'url': "https://aip.baidubce.com/oauth/2.0/token",
        >>>     'url_htp': 'get',
        >>>     'prm': {
        >>>         'grant_type': 'client_credentials',
        >>>         'client_id': 'HPn***uG8',
        >>>         'client_secret': 'QLa***Pmj',
        >>>     },
        >>>     'ppc': {
        >>>         'key': ['access_token'],    # key name for access token
        >>>         'xpr': 2592000,             # expire in some seconds
        >>>     }
        >>> }   # 以百度文字识别token为例
        @param dly: delay seconds, default 259200 for baiduMap and 3600 for wechat
        @param frc: force to update token
        @param rty: max retry times
        @return: token in type str
        """
        dly = self.lcn['ppc']['xpr'] if 'xpr' in self.lcn['ppc'].keys() else dly
        prc, bch = True, 0
        if os_exists(os_join(self.lcn['fld'], self.lcn['fls'])):
            self.mpt_txt()          # 当目录中已经存在txt文件时, 读取
            self.dts = loads(self.dts[0].replace('\n', '').replace("'", '"'))
        else:                       # 否则虚拟一个必定超时的时间点
            self.dts = {'tms': '2020-01-01 00:00:01'}
            if not os_exists(self.lcn['fld']):
                makedirs(self.lcn['fld'])
        while prc and bch <= rty:
            if frc or dtz('now').val - dtz(self.dts['tms']).typ_to_dtt(rtn=True) >= typ_dt_timedelta(seconds=dly):
                self.api_run()      # 当强制刷新参数为True或超时时重新调取token
                self.dts.update({'tms': dtz('now').dtt_to_typ(str_fmt='%Y-%m-%d %H:%M:%S', rtn=True)})
                self.lcl_xpt()      # 生成本地存放文件
                frc = False
            else:
                return self.dts[self.lcn['ppc']['key'][0]]
            bch += 1


class ioz(mngMixin, sqlMixin, apiMixin, lclMixin):
    """
    input/output class on multiple methods.
    """
    def __init__(self, dts=None, lcn=None, *, spr=False):
        super(ioz, self).__init__(dts, lcn, spr=spr)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on 2021.03.19
A package slot for modelling
requirements = [
    'matplotlib>=3.1.3',
    'sklearn>=0.22.1',
    'seaborn>=0.10.0',
    'factor_analyzer>=0.3.2',
    'joblib>=0.14.1',
]
@author: zoharslong
"""
from pyzohar.sub_slt_mdl.mdz import mdz

__all__ = ['mdz']
print('info: modelling slot inserted.')
